# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import os
import platform
from unittest import mock
import time
import unittest

from azure.mgmt.core.tools import parse_resource_id

from azure.cli.command_modules.containerapp.custom import containerapp_ssh
from azure.cli.command_modules.containerapp.tests.latest.utils import create_containerapp_env, \
    prepare_containerapp_env_for_app_e2e_tests
from azure.cli.core.azclierror import InvalidArgumentValueError

from azure.cli.testsdk.reverse_dependency import get_dummy_cli
from azure.cli.testsdk.scenario_tests import AllowLargeResponse
from azure.cli.testsdk import (ScenarioTest, ResourceGroupPreparer, JMESPathCheck, StringCheck, live_only,
                               JMESPathCheckNotExists, LogAnalyticsWorkspacePreparer)
from knack.util import CLIError

from azure.cli.command_modules.containerapp.tests.latest.common import TEST_LOCATION, write_test_file, \
    clean_up_test_file

TEST_DIR = os.path.abspath(os.path.join(os.path.abspath(__file__), '..'))
# flake8: noqa
# noqa
# pylint: skip-file


class ContainerappScenarioTest(ScenarioTest):
    def __init__(self, *arg, **kwargs):
        super().__init__(*arg, random_config_dir=True, **kwargs)

    @AllowLargeResponse(8192)
    @ResourceGroupPreparer(location="eastus2")
    def test_containerapp_e2e(self, resource_group):
        self.cmd('configure --defaults location={}'.format(TEST_LOCATION))

        env = prepare_containerapp_env_for_app_e2e_tests(self)

        containerapp_name = self.create_random_name(prefix='containerapp-e2e', length=24)

        # Create basic Container App with default image
        self.cmd('containerapp create -g {} -n {} --environment {}'.format(resource_group, containerapp_name, env), checks=[
            JMESPathCheck('name', containerapp_name)
        ])

        self.cmd('containerapp show -g {} -n {}'.format(resource_group, containerapp_name), checks=[
            JMESPathCheck('name', containerapp_name),
        ])

        self.cmd('containerapp list -g {}'.format(resource_group), checks=[
            JMESPathCheck('length(@)', 1),
            JMESPathCheck('[0].name', containerapp_name),
        ])

        # Create Container App with image, resource and replica limits
        create_string = "containerapp create -g {} -n {} --environment {} --image mcr.microsoft.com/azuredocs/containerapps-helloworld:latest --cpu 0.5 --memory 1.0Gi --min-replicas 2 --max-replicas 4".format(resource_group, containerapp_name, env)
        self.cmd(create_string, checks=[
            JMESPathCheck('name', containerapp_name),
            JMESPathCheck('properties.template.containers[0].image', 'mcr.microsoft.com/azuredocs/containerapps-helloworld:latest'),
            JMESPathCheck('properties.template.containers[0].resources.cpu', '0.5'),
            JMESPathCheck('properties.template.containers[0].resources.memory', '1Gi'),
            JMESPathCheck('properties.template.scale.minReplicas', '2'),
            JMESPathCheck('properties.template.scale.maxReplicas', '4')
        ])

        self.cmd('containerapp create -g {} -n {} --environment {} --ingress external --target-port 8080'.format(resource_group, containerapp_name, env), checks=[
            JMESPathCheck('properties.configuration.ingress.external', True),
            JMESPathCheck('properties.configuration.ingress.targetPort', 8080)
        ])

        # target port is optional
        self.cmd('containerapp create -g {} -n {} --environment {} --ingress external'.format(resource_group, containerapp_name, env), checks=[
            JMESPathCheck('properties.configuration.ingress.external', True),
            JMESPathCheck('properties.configuration.ingress.targetPort', 0)
        ])

        # Create Container App with secrets and environment variables
        containerapp_name = self.create_random_name(prefix='containerapp-e2e', length=24)
        create_string = 'containerapp create -g {} -n {} --environment {} --secrets mysecret=secretvalue1 anothersecret="secret value 2" --env-vars GREETING="Hello, world" SECRETENV=secretref:anothersecret'.format(
            resource_group, containerapp_name, env)
        self.cmd(create_string, checks=[
            JMESPathCheck('name', containerapp_name),
            JMESPathCheck('length(properties.template.containers[0].env)', 2),
            JMESPathCheck('length(properties.configuration.secrets)', 2)
        ])

    @live_only()  # Pass lively, But failed in playback mode with error: WebSocketBadStatusException: Handshake status 401 Unauthorized
    @ResourceGroupPreparer(location="eastus2")
    def test_containerapp_exec(self, resource_group):
        self.cmd('configure --defaults location={}'.format(TEST_LOCATION))
        env = prepare_containerapp_env_for_app_e2e_tests(self)

        containerapp_name = self.create_random_name(prefix='containerapp-e2e1', length=24)
        # create an app with ingress is None
        app = self.cmd(f'containerapp create -g {resource_group} -n {containerapp_name} --environment {env}', checks=[
            JMESPathCheck('name', containerapp_name),
            JMESPathCheck('properties.configuration.ingress', None),

        ]).get_output_in_json()

        self.containerapp_exec_test_helper(resource_group, containerapp_name, app["properties"]["latestRevisionName"])

        #  Test external App
        external_containerapp_name = self.create_random_name(prefix='containerapp-e2e2', length=24)
        # create an app with ingress is None
        external_containerapp = self.cmd(f'containerapp create -g {resource_group} -n {external_containerapp_name} --environment {env} --ingress external --target-port 8080', checks=[
            JMESPathCheck('name', external_containerapp_name),
            JMESPathCheck('properties.configuration.ingress.external', True),
            JMESPathCheck('properties.configuration.ingress.targetPort', 8080)
        ]).get_output_in_json()

        self.containerapp_exec_test_helper(resource_group, external_containerapp_name, external_containerapp["properties"]["latestRevisionName"])

        #  Test internal App
        # update ingress to internal
        self.cmd(
            f'containerapp ingress update -g {resource_group} -n {external_containerapp_name} --type internal',
            checks=[
                JMESPathCheck('external', False),
            ]).get_output_in_json()
        internal_containerapp = self.cmd(f'containerapp show -g {resource_group} -n {external_containerapp_name}').get_output_in_json()
        self.containerapp_exec_test_helper(resource_group, external_containerapp_name, internal_containerapp["properties"]["latestRevisionName"])

    def containerapp_exec_test_helper(self, resource_group, containerapp_name, latest_revision_name):
        self.cmd(f'containerapp exec -g {resource_group} -n {containerapp_name}')

        self.cmd(f'containerapp exec -g {resource_group} -n {containerapp_name} --command ls')

        replica_list = self.cmd(
            f'containerapp replica list -g {resource_group} -n {containerapp_name} --revision {latest_revision_name}',
            expect_failure=False).get_output_in_json()

        self.cmd(
            f'containerapp exec -g {resource_group} -n {containerapp_name} --replica {replica_list[0]["name"]} --revision {latest_revision_name} --command ls',
            expect_failure=False)

    @AllowLargeResponse(8192)
    @ResourceGroupPreparer(location="eastus2")
    @LogAnalyticsWorkspacePreparer(location="eastus", get_shared_key=True)
    def test_container_acr(self, resource_group, laworkspace_customer_id, laworkspace_shared_key):
        self.cmd('configure --defaults location={}'.format(TEST_LOCATION))

        env_id = prepare_containerapp_env_for_app_e2e_tests(self)

        containerapp_name = self.create_random_name(prefix='containerapp-e2e', length=24)
        registry_name = self.create_random_name(prefix='containerapp', length=24)

        # Create ACR
        acr = self.cmd('acr create -g {} -n {} --sku Basic --admin-enabled'.format(resource_group, registry_name)).get_output_in_json()
        registry_server = acr["loginServer"]

        acr_credentials = self.cmd('acr credential show -g {} -n {}'.format(resource_group, registry_name)).get_output_in_json()
        registry_username = acr_credentials["username"]
        registry_password = acr_credentials["passwords"][0]["value"]

        # Create Container App with ACR
        containerapp_name = self.create_random_name(prefix='containerapp-e2e', length=24)
        create_string = 'containerapp create -g {} -n {} --environment {} --registry-username {} --registry-server {} --registry-password {}'.format(
            resource_group, containerapp_name, env_id, registry_username, registry_server, registry_password)
        self.cmd(create_string, checks=[
            JMESPathCheck('name', containerapp_name),
            JMESPathCheck('properties.configuration.registries[0].server', registry_server),
            JMESPathCheck('properties.configuration.registries[0].username', registry_username),
            JMESPathCheck('length(properties.configuration.secrets)', 1),
        ])


    # TODO rename
    @AllowLargeResponse(8192)
    @ResourceGroupPreparer(location="westeurope")
    def test_containerapp_update(self, resource_group):
        self.cmd('configure --defaults location={}'.format(TEST_LOCATION))
        env_id = prepare_containerapp_env_for_app_e2e_tests(self)

        # Create basic Container App with default image
        containerapp_name = self.create_random_name(prefix='containerapp-update', length=24)

        self.cmd('containerapp create -g {} -n {} --environment {}'.format(resource_group, containerapp_name, env_id), checks=[
            JMESPathCheck('name', containerapp_name),
            JMESPathCheck('length(properties.template.containers)', 1),
            JMESPathCheck('properties.template.containers[0].name', containerapp_name)
        ])

        self.cmd('containerapp show -g {} -n {}'.format(resource_group, containerapp_name), checks=[
            JMESPathCheck('name', containerapp_name),
        ])

        self.cmd('containerapp list -g {}'.format(resource_group), checks=[
            JMESPathCheck('length(@)', 1),
            JMESPathCheck('[0].name', containerapp_name),
        ])

        # Create Container App with image, resource and replica limits
        create_string = "containerapp create -g {} -n {} --environment {} --image mcr.microsoft.com/azuredocs/containerapps-helloworld:latest --cpu 0.5 --memory 1.0Gi --min-replicas 2 --max-replicas 4".format(resource_group, containerapp_name, env_id)
        self.cmd(create_string, checks=[
            JMESPathCheck('name', containerapp_name),
            JMESPathCheck('properties.template.containers[0].image', 'mcr.microsoft.com/azuredocs/containerapps-helloworld:latest'),
            JMESPathCheck('properties.template.containers[0].resources.cpu', '0.5'),
            JMESPathCheck('properties.template.containers[0].resources.memory', '1Gi'),
            JMESPathCheck('properties.template.scale.minReplicas', '2'),
            JMESPathCheck('properties.template.scale.maxReplicas', '4')
        ])

        self.cmd('containerapp create -g {} -n {} --environment {} --ingress external --target-port 8080'.format(resource_group, containerapp_name, env_id), checks=[
            JMESPathCheck('properties.configuration.ingress.external', True),
            JMESPathCheck('properties.configuration.ingress.targetPort', 8080)
        ])

        # target port is optional
        self.cmd('containerapp create -g {} -n {} --environment {} --ingress external'.format(resource_group, containerapp_name, env_id), checks=[
            JMESPathCheck('properties.configuration.ingress.external', True),
            JMESPathCheck('properties.configuration.ingress.targetPort', 0)
        ])

        self.cmd('containerapp ingress disable -g {} -n {}'.format(resource_group, containerapp_name), checks=[
            StringCheck('')
        ])

        self.cmd('containerapp show -g {} -n {}'.format(resource_group, containerapp_name), checks=[
            JMESPathCheckNotExists('properties.configuration.ingress'),
        ])

        self.cmd('containerapp ingress enable -g {} -n {} --type external --target-port 8080'.format(resource_group, containerapp_name), checks=[
            JMESPathCheck('external', True),
            JMESPathCheck('targetPort', 8080)
        ])

        self.cmd('containerapp ingress disable -g {} -n {}'.format(resource_group, containerapp_name), checks=[
            StringCheck('')
        ])

        self.cmd('containerapp show -g {} -n {}'.format(resource_group, containerapp_name), checks=[
            JMESPathCheckNotExists('properties.configuration.ingress'),
        ])

        self.cmd('containerapp ingress enable -g {} -n {} --type external'.format(resource_group, containerapp_name), checks=[
            JMESPathCheck('external', True),
            JMESPathCheck('targetPort', 0)
        ])

        # Create Container App with secrets and environment variables
        containerapp_name = self.create_random_name(prefix='containerapp-e2e', length=24)
        create_string = 'containerapp create -g {} -n {} --environment {} --secrets mysecret=secretvalue1 anothersecret="secret value 2" --env-vars GREETING="Hello, world" SECRETENV=secretref:anothersecret'.format(
            resource_group, containerapp_name, env_id)
        self.cmd(create_string, checks=[
            JMESPathCheck('name', containerapp_name),
            JMESPathCheck('length(properties.template.containers[0].env)', 2),
            JMESPathCheck('length(properties.configuration.secrets)', 2)
        ])


    @AllowLargeResponse(8192)
    @ResourceGroupPreparer(location="eastus2")
    def test_container_acr_env_var(self, resource_group):
        self.cmd('configure --defaults location={}'.format(TEST_LOCATION))

        env = prepare_containerapp_env_for_app_e2e_tests(self)

        registry_name = self.create_random_name(prefix='containerapp', length=24)

        # Create ACR
        acr = self.cmd('acr create -g {} -n {} --sku Basic --admin-enabled'.format(resource_group, registry_name)).get_output_in_json()
        registry_server = acr["loginServer"]

        acr_credentials = self.cmd('acr credential show -g {} -n {}'.format(resource_group, registry_name)).get_output_in_json()
        registry_username = acr_credentials["username"]
        registry_password = acr_credentials["passwords"][0]["value"]

        # Create Container App with ACR
        containerapp_name = self.create_random_name(prefix='containerapp-e2e', length=24)
        create_string = 'containerapp create -g {} -n {} --environment {} --registry-username {} --registry-server {} --registry-password {}'.format(
            resource_group, containerapp_name, env, registry_username, registry_server, registry_password)
        self.cmd(create_string, checks=[
            JMESPathCheck('name', containerapp_name),
            JMESPathCheck('properties.configuration.registries[0].server', registry_server),
            JMESPathCheck('properties.configuration.registries[0].username', registry_username),
            JMESPathCheck('length(properties.configuration.secrets)', 1),
        ])

        # Update Container App with ACR, set --min-replicas 1
        update_string = 'containerapp update -g {} -n {} --min-replicas 1 --max-replicas 1 --set-env-vars testenv=testing'.format(
            resource_group, containerapp_name)
        self.cmd(update_string, checks=[
            JMESPathCheck('name', containerapp_name),
            JMESPathCheck('properties.configuration.registries[0].server', registry_server),
            JMESPathCheck('properties.configuration.registries[0].username', registry_username),
            JMESPathCheck('length(properties.configuration.secrets)', 1),
            JMESPathCheck('properties.template.scale.minReplicas', '1'),
            JMESPathCheck('properties.template.scale.maxReplicas', '1'),
            JMESPathCheck('length(properties.template.containers[0].env)', 1),
            JMESPathCheck('properties.template.containers[0].env[0].name', "testenv"),
            JMESPathCheck('properties.template.containers[0].env[0].value', None),
        ])

        # Add secrets to Container App with ACR
        containerapp_secret = self.cmd('containerapp secret list -g {} -n {}'.format(resource_group, containerapp_name)).get_output_in_json()
        secret_name = containerapp_secret[0]["name"]
        secret_string = 'containerapp secret set -g {} -n {} --secrets newsecret=test'.format(resource_group, containerapp_name)
        self.cmd(secret_string, checks=[
            JMESPathCheck('length(@)', 2),
        ])

        self.cmd('containerapp secret show -g {} -n {} --secret-name {}'.format(resource_group, containerapp_name, secret_name), checks=[
            JMESPathCheck('name', secret_name),
        ])

        with self.assertRaises(CLIError):
            # Removing ACR password should fail since it is needed for ACR
            self.cmd('containerapp secret remove -g {} -n {} --secret-names {}'.format(resource_group, containerapp_name, secret_name))

        # Update Container App with ACR, --min-replicas 0
        update_string = 'containerapp update -g {} -n {} --min-replicas 0 --max-replicas 1 --replace-env-vars testenv=testing2'.format(
            resource_group, containerapp_name)
        self.cmd(update_string, checks=[
            JMESPathCheck('name', containerapp_name),
            JMESPathCheck('properties.configuration.registries[0].server', registry_server),
            JMESPathCheck('properties.configuration.registries[0].username', registry_username),
            JMESPathCheck('properties.template.scale.minReplicas', '0'),
            JMESPathCheck('properties.template.scale.maxReplicas', '1'),
            JMESPathCheck('length(properties.template.containers[0].env)', 1),
            JMESPathCheck('properties.template.containers[0].env[0].name', "testenv"),
            JMESPathCheck('properties.template.containers[0].env[0].value', None),
        ])
        update_string = 'containerapp update -g {} -n {} --min-replicas 0 --max-replicas 1 --remove-env-vars testenv --remove-all-env-vars'.format(resource_group, containerapp_name)
        self.cmd(update_string, checks=[
            JMESPathCheck('name', containerapp_name),
            JMESPathCheck('properties.configuration.registries[0].server', registry_server),
            JMESPathCheck('properties.configuration.registries[0].username', registry_username),
            JMESPathCheck('properties.template.scale.minReplicas', '0'),
            JMESPathCheck('properties.template.scale.maxReplicas', '1'),
            JMESPathCheck('properties.template.containers[0].env', None),
        ])

    @AllowLargeResponse(8192)
    @ResourceGroupPreparer(location="northeurope")
    @LogAnalyticsWorkspacePreparer(location="eastus", get_shared_key=True)
    def test_containerapp_update_containers(self, resource_group):
        self.cmd('configure --defaults location={}'.format(TEST_LOCATION))

        env = prepare_containerapp_env_for_app_e2e_tests(self)

        # Create basic Container App with default image
        containerapp_name = self.create_random_name(prefix='containerapp-update', length=24)
        self.cmd('containerapp create -g {} -n {} --environment {}'.format(resource_group, containerapp_name, env), checks=[
            JMESPathCheck('name', containerapp_name),
            JMESPathCheck('length(properties.template.containers)', 1),
            JMESPathCheck('properties.template.containers[0].name', containerapp_name)
        ])

        # Update existing Container App that has a single container

        update_string = 'containerapp update -g {} -n {} --image {} --cpu 0.5 --memory 1.0Gi --args mycommand mycommand2 --command "mycommand" --revision-suffix suffix --min-replicas 2 --max-replicas 34'.format(
            resource_group, containerapp_name, 'nginx')
        self.cmd(update_string, checks=[
            JMESPathCheck('name', containerapp_name),
            JMESPathCheck('length(properties.template.containers)', 1),
            JMESPathCheck('properties.template.containers[0].name', containerapp_name),
            JMESPathCheck('properties.template.containers[0].image', 'nginx'),
            JMESPathCheck('properties.template.containers[0].resources.cpu', '0.5'),
            JMESPathCheck('properties.template.containers[0].resources.memory', '1Gi'),
            JMESPathCheck('properties.template.scale.minReplicas', '2'),
            JMESPathCheck('properties.template.scale.maxReplicas', '34'),
            JMESPathCheck('properties.template.containers[0].command[0]', "mycommand"),
            JMESPathCheck('length(properties.template.containers[0].args)', 2)
        ])

        # Add new container to existing Container App
        update_string = 'containerapp update -g {} -n {} --container-name {} --image {}'.format(
            resource_group, containerapp_name, "newcontainer", "nginx")
        self.cmd(update_string, checks=[
            JMESPathCheck('name', containerapp_name),
            JMESPathCheck('length(properties.template.containers)', 2)
        ])

        # Updating container properties in a Container App with multiple containers, without providing container name should error
        update_string = 'containerapp update -g {} -n {} --cpu {} --memory {}'.format(
            resource_group, containerapp_name, '1.0', '2.0Gi')
        with self.assertRaises(CLIError):
            self.cmd(update_string)

        # Updating container properties in a Container App with multiple containers, should work when container name provided
        update_string = 'containerapp update -g {} -n {} --container-name {} --cpu {} --memory {}'.format(
            resource_group, containerapp_name, 'newcontainer', '0.75', '1.5Gi')
        self.cmd(update_string)

        update_string = 'containerapp update -g {} -n {} --container-name {} --cpu {} --memory {}'.format(
            resource_group, containerapp_name, containerapp_name, '0.75', '1.5Gi')
        self.cmd(update_string, checks=[
            JMESPathCheck('properties.template.containers[0].resources.cpu', '0.75'),
            JMESPathCheck('properties.template.containers[0].resources.memory', '1.5Gi'),
            JMESPathCheck('properties.template.containers[1].resources.cpu', '0.75'),
            JMESPathCheck('properties.template.containers[1].resources.memory', '1.5Gi'),
        ])

    # TODO fix and enable
    @unittest.skip("API only on stage currently")
    @live_only()  # VCR.py can't seem to handle websockets (only --live works)
    # @ResourceGroupPreparer(location="centraluseuap")
    @mock.patch("azure.cli.command_modules.containerapp._ssh_utils._resize_terminal")
    @mock.patch("sys.stdin")
    def test_containerapp_ssh(self, resource_group=None, *args):
        self.cmd('configure --defaults location={}'.format(TEST_LOCATION))

        # containerapp_name = self.create_random_name(prefix='capp', length=24)
        # env_name = self.create_random_name(prefix='env', length=24)

        # self.cmd(f'containerapp env create -g {resource_group} -n {env_name}')
        # self.cmd(f'containerapp create -g {resource_group} -n {containerapp_name} --environment {env_name} --min-replicas 1 --ingress external')

        # TODO remove hardcoded app info (currently the SSH feature is only enabled in stage)
        # these are only in my sub so they won't work on the CI / other people's machines
        containerapp_name = "stage"
        resource_group = "sca"

        stdout_buff = []

        def mock_print(*args, end="\n", **kwargs):
            out = " ".join([str(a) for a in args])
            if not stdout_buff:
                stdout_buff.append(out)
            elif end != "\n":
                stdout_buff[-1] = f"{stdout_buff[-1]}{out}"
            else:
                stdout_buff.append(out)

        commands = "\n".join(["whoami", "pwd", "ls -l | grep index.js", "exit\n"])
        expected_output = ["root", "/usr/src/app", "-rw-r--r--    1 root     root           267 Oct 15 00:21 index.js"]

        idx = [0]
        def mock_getch():
            ch = commands[idx[0]].encode("utf-8")
            idx[0] = (idx[0] + 1) % len(commands)
            return ch

        cmd = mock.MagicMock()
        cmd.cli_ctx = get_dummy_cli()
        from azure.cli.command_modules.containerapp._validators import validate_ssh
        from azure.cli.command_modules.containerapp.custom import containerapp_ssh

        class Namespace: pass
        namespace = Namespace()
        setattr(namespace, "name", containerapp_name)
        setattr(namespace, "resource_group_name", resource_group)
        setattr(namespace, "revision", None)
        setattr(namespace, "replica", None)
        setattr(namespace, "container", None)

        validate_ssh(cmd=cmd, namespace=namespace)  # needed to set values for container, replica, revision

        mock_lib = "tty.setcbreak"
        if platform.system() == "Windows":
            mock_lib = "azure.cli.command_modules.containerapp._ssh_utils.enable_vt_mode"

        with mock.patch("builtins.print", side_effect=mock_print), mock.patch(mock_lib):
            with mock.patch("azure.cli.command_modules.containerapp._ssh_utils._getch_unix", side_effect=mock_getch), mock.patch("azure.cli.command_modules.containerapp._ssh_utils._getch_windows", side_effect=mock_getch):
                containerapp_ssh(cmd=cmd, resource_group_name=namespace.resource_group_name, name=namespace.name,
                                    container=namespace.container, revision=namespace.revision, replica=namespace.replica, startup_command="sh")
        for line in expected_output:
            self.assertIn(line, expected_output)

    @ResourceGroupPreparer(location="northeurope")
    def test_containerapp_logstream(self, resource_group):
        containerapp_name = self.create_random_name(prefix='capp', length=24)
        env = prepare_containerapp_env_for_app_e2e_tests(self)

        self.cmd(f'containerapp create -g {resource_group} -n {containerapp_name} --environment {env} --min-replicas 1 --ingress external --target-port 80')

        self.cmd(f'containerapp logs show -n {containerapp_name} -g {resource_group} --follow false --format json --tail 10', expect_failure=False)

    @ResourceGroupPreparer(location="northeurope")
    def test_containerapp_eventstream(self, resource_group):
        self.cmd('configure --defaults location={}'.format(TEST_LOCATION))

        containerapp_name = self.create_random_name(prefix='capp', length=24)
        env_id = prepare_containerapp_env_for_app_e2e_tests(self)
        env_rg = parse_resource_id(env_id).get('resource_group')
        env_name = parse_resource_id(env_id).get('name')

        self.cmd(f'containerapp create -g {resource_group} -n {containerapp_name} --environment {env_id} --min-replicas 1 --ingress external --target-port 80')

        self.cmd(f'containerapp logs show -n {containerapp_name} -g {resource_group} --type system')
        self.cmd(f'containerapp env logs show -n {env_name} -g {env_rg} --tail 15 --follow false')

    @live_only()  # Pass lively, But failed in playback mode with Role assignment error, which is expected: azure.cli.core.azclierror.UnauthorizedError: Role assignment failed with error
    @ResourceGroupPreparer(location="northeurope")
    @AllowLargeResponse(size_kb=99999)
    def test_containerapp_registry_msi(self, resource_group):
        self.cmd('configure --defaults location={}'.format(TEST_LOCATION))

        app = self.create_random_name(prefix='app', length=24)
        acr = self.create_random_name(prefix='acr', length=24)
        env = prepare_containerapp_env_for_app_e2e_tests(self)

        self.cmd(f'acr create -g {resource_group} -n {acr} --sku basic --admin-enabled')

        # This command will create arcpull for the system-assigned containerapp
        self.cmd(f'containerapp create -g {resource_group} -n {app} --environment {env} --min-replicas 1 --ingress external --target-port 80 --system-assigned --registry-server {acr}.azurecr.io --registry-identity system', checks=[
            JMESPathCheck("properties.provisioningState", "Succeeded"),
            JMESPathCheck("properties.configuration.registries[0].passwordSecretRef", ""),
            JMESPathCheck("properties.configuration.registries[0].username", ""),
            JMESPathCheck("properties.configuration.registries[0].identity", "system"),
        ])
        self.cmd(f'containerapp registry remove -g {resource_group} -n {app} --server {acr}.azurecr.io',expect_failure=False)
        self.cmd(f'containerapp show -g {resource_group} -n {app}', checks=[
            JMESPathCheck("properties.provisioningState", "Succeeded"),
            JMESPathCheck("properties.configuration.registries", None),
        ])

        # self.cmd(f'acr credential renew -n {acr} ')
        self.cmd(f'containerapp registry set --server {acr}.azurecr.io -g {resource_group} -n {app}')
        registry_list = self.cmd(f'containerapp registry list -g {resource_group} -n {app}', checks=[
            JMESPathCheck('length(@)', 1),
        ]).get_output_in_json()

        self.cmd(f'containerapp registry show -g {resource_group} -n {app} --server {registry_list[0]["server"]}', checks=[
            JMESPathCheck('server', acr+'.azurecr.io'),
        ])

        with self.assertRaises(InvalidArgumentValueError):
            self.cmd(f'containerapp registry show -g {resource_group} -n {app} --server abc')

        app_data = self.cmd(f'containerapp show -g {resource_group} -n {app}', checks=[
            JMESPathCheck("properties.provisioningState", "Succeeded"),
            JMESPathCheck("properties.configuration.registries[0].server", f'{acr}.azurecr.io'),
        ]).get_output_in_json()
        self.assertIsNotNone(app_data["properties"]["configuration"]["registries"][0].get("passwordSecretRef"))
        self.assertIsNotNone(app_data["properties"]["configuration"]["registries"][0].get("username"))
        self.assertEqual(app_data["properties"]["configuration"]["registries"][0].get("identity"), "")

        self.cmd(f'containerapp registry set --server {acr}.azurecr.io -g {resource_group} -n {app} --identity system')
        app_data = self.cmd(f'containerapp show -g {resource_group} -n {app}', checks=[
            JMESPathCheck("properties.provisioningState", "Succeeded"),
            JMESPathCheck("properties.configuration.registries[0].server", f'{acr}.azurecr.io'),
            JMESPathCheck("properties.configuration.registries[0].passwordSecretRef", ""),
            JMESPathCheck("properties.configuration.registries[0].username", ""),
            JMESPathCheck("properties.configuration.registries[0].identity", "system"),
        ]).get_output_in_json()

        self.cmd(f'containerapp registry remove -g {resource_group} -n {app} --server {registry_list[0]["server"]}',expect_failure=False)

    @AllowLargeResponse(8192)
    @ResourceGroupPreparer(location="westeurope")
    @LogAnalyticsWorkspacePreparer(location="eastus", get_shared_key=True)
    def test_containerapp_create_with_environment_id(self, resource_group, laworkspace_customer_id, laworkspace_shared_key):
        self.cmd('configure --defaults location={}'.format(TEST_LOCATION))

        env1 = self.create_random_name(prefix='env1', length=24)
        env2 = self.create_random_name(prefix='env2', length=24)

        app = self.create_random_name(prefix='yaml1', length=24)

        create_containerapp_env(self, env1, resource_group, logs_workspace=laworkspace_customer_id, logs_workspace_shared_key=laworkspace_shared_key)
        containerapp_env1 = self.cmd(
            'containerapp env show -g {} -n {}'.format(resource_group, env1)).get_output_in_json()

        create_containerapp_env(self, env2, resource_group, logs_workspace=laworkspace_customer_id, logs_workspace_shared_key=laworkspace_shared_key)
        containerapp_env2 = self.cmd(
            'containerapp env show -g {} -n {}'.format(resource_group, env2)).get_output_in_json()
        # test `az containerapp up` with --environment
        image = 'mcr.microsoft.com/azuredocs/aks-helloworld:v1'
        ca_name = self.create_random_name(prefix='containerapp', length=24)
        self.cmd('containerapp up -g {} -n {} --environment {} --image {}'.format(resource_group, ca_name, env2, image), expect_failure=False)
        self.cmd(f'containerapp show -g {resource_group} -n {ca_name}', checks=[
            JMESPathCheck("properties.provisioningState", "Succeeded"),
            JMESPathCheck("properties.environmentId", containerapp_env2["id"]),
            JMESPathCheck('properties.template.containers[0].image', image),
        ])
        # test `az containerapp up` for existing containerapp without --environment
        image2 = 'mcr.microsoft.com/k8se/quickstart:latest'
        self.cmd('containerapp up -g {} -n {} --image {}'.format(resource_group, ca_name, image2), expect_failure=False)
        self.cmd(f'containerapp show -g {resource_group} -n {ca_name}', checks=[
            JMESPathCheck("properties.provisioningState", "Succeeded"),
            JMESPathCheck("properties.environmentId", containerapp_env2["id"]),
            JMESPathCheck('properties.template.containers[0].image', image2),
        ])

        user_identity_name = self.create_random_name(prefix='containerapp-user', length=24)
        user_identity = self.cmd(
            'identity create -g {} -n {}'.format(resource_group, user_identity_name)).get_output_in_json()
        user_identity_id = user_identity['id']

        # the value in --yaml is used, warning for different value in --environmentId
        containerapp_yaml_text = f"""
                                    location: {TEST_LOCATION}
                                    type: Microsoft.App/containerApps
                                    tags:
                                        tagname: value
                                    properties:
                                      environmentId: {containerapp_env1["id"]}
                                      configuration:
                                        activeRevisionsMode: Multiple
                                        ingress:
                                          external: false
                                          allowInsecure: false
                                          targetPort: 80
                                          traffic:
                                            - latestRevision: true
                                              weight: 100
                                          transport: Auto
                                          ipSecurityRestrictions:
                                            - name: name
                                              ipAddressRange: "1.1.1.1/10"
                                              action: "Allow"
                                      template:
                                        revisionSuffix: myrevision
                                        terminationGracePeriodSeconds: 90
                                        containers:
                                          - image: nginx
                                            name: nginx
                                            env:
                                              - name: HTTP_PORT
                                                value: 80
                                            command:
                                              - npm
                                              - start
                                            resources:
                                              cpu: 0.5
                                              memory: 1Gi
                                        scale:
                                          minReplicas: 1
                                          maxReplicas: 3
                                          rules:
                                          - http:
                                              auth:
                                              - secretRef: secretref
                                                triggerParameter: trigger
                                              metadata:
                                                concurrentRequests: '50'
                                                key: value
                                            name: http-scale-rule
                                    identity:
                                      type: UserAssigned
                                      userAssignedIdentities:
                                        {user_identity_id}: {{}}
                                    """
        containerapp_file_name = f"{self._testMethodName}_containerapp.yml"

        write_test_file(containerapp_file_name, containerapp_yaml_text)
        self.cmd(
            f'containerapp create -n {app} -g {resource_group} --environment {env2} --yaml {containerapp_file_name}')

        self.cmd(f'containerapp show -g {resource_group} -n {app}', checks=[
            JMESPathCheck("properties.provisioningState", "Succeeded"),
            JMESPathCheck("properties.environmentId", containerapp_env1["id"]),
            JMESPathCheck("properties.configuration.ingress.external", False),
            JMESPathCheck("properties.configuration.ingress.ipSecurityRestrictions[0].name", "name"),
            JMESPathCheck("properties.configuration.ingress.ipSecurityRestrictions[0].ipAddressRange",
                          "1.1.1.1/10"),
            JMESPathCheck("properties.configuration.ingress.ipSecurityRestrictions[0].action", "Allow"),
            JMESPathCheck("properties.environmentId", containerapp_env1["id"]),
            JMESPathCheck("properties.template.revisionSuffix", "myrevision"),
            JMESPathCheck("properties.template.terminationGracePeriodSeconds", 90),
            JMESPathCheck("properties.template.containers[0].name", "nginx"),
            JMESPathCheck("properties.template.scale.minReplicas", 1),
            JMESPathCheck("properties.template.scale.maxReplicas", 3),
            JMESPathCheck("properties.template.scale.rules[0].name", "http-scale-rule"),
            JMESPathCheck("properties.template.scale.rules[0].http.metadata.concurrentRequests", "50"),
            JMESPathCheck("properties.template.scale.rules[0].http.metadata.key", "value"),
            JMESPathCheck("properties.template.scale.rules[0].http.auth[0].triggerParameter", "trigger"),
            JMESPathCheck("properties.template.scale.rules[0].http.auth[0].secretRef", "secretref"),
        ])

        containerapp_yaml_text = f"""
                                            location: {TEST_LOCATION}
                                            type: Microsoft.App/containerApps
                                            tags:
                                                tagname: value
                                            properties:
                                              configuration:
                                                activeRevisionsMode: Multiple
                                                ingress:
                                                  external: false
                                                  allowInsecure: false
                                                  targetPort: 80
                                                  traffic:
                                                    - latestRevision: true
                                                      weight: 100
                                                  transport: Auto
                                                  ipSecurityRestrictions:
                                                    - name: name
                                                      ipAddressRange: "1.1.1.1/10"
                                                      action: "Allow"
                                              template:
                                                revisionSuffix: myrevision
                                                terminationGracePeriodSeconds: 90
                                                containers:
                                                  - image: nginx
                                                    name: nginx
                                                    env:
                                                      - name: HTTP_PORT
                                                        value: 80
                                                    command:
                                                      - npm
                                                      - start
                                                    resources:
                                                      cpu: 0.5
                                                      memory: 1Gi
                                                scale:
                                                  minReplicas: 1
                                                  maxReplicas: 3
                                                  rules:
                                                  - http:
                                                      auth:
                                                      - secretRef: secretref
                                                        triggerParameter: trigger
                                                      metadata:
                                                        concurrentRequests: '50'
                                                        key: value
                                                    name: http-scale-rule
                                            identity:
                                              type: UserAssigned
                                              userAssignedIdentities:
                                                {user_identity_id}: {{}}
                                            """

        write_test_file(containerapp_file_name, containerapp_yaml_text)
        app2 = self.create_random_name(prefix='yaml2', length=24)
        self.cmd(
            f'containerapp create -n {app2} -g {resource_group} --environment {env2} --yaml {containerapp_file_name}')
        self.cmd(f'containerapp show -g {resource_group} -n {app2}', checks=[
            JMESPathCheck("properties.provisioningState", "Succeeded"),
            JMESPathCheck("properties.environmentId", containerapp_env2["id"]),
            JMESPathCheck("properties.configuration.ingress.external", False),
            JMESPathCheck("properties.configuration.ingress.ipSecurityRestrictions[0].name", "name"),
            JMESPathCheck("properties.configuration.ingress.ipSecurityRestrictions[0].ipAddressRange",
                          "1.1.1.1/10"),
            JMESPathCheck("properties.configuration.ingress.ipSecurityRestrictions[0].action", "Allow"),
            JMESPathCheck("properties.environmentId", containerapp_env2["id"]),
            JMESPathCheck("properties.template.revisionSuffix", "myrevision"),
            JMESPathCheck("properties.template.terminationGracePeriodSeconds", 90),
            JMESPathCheck("properties.template.containers[0].name", "nginx"),
            JMESPathCheck("properties.template.scale.minReplicas", 1),
            JMESPathCheck("properties.template.scale.maxReplicas", 3),
            JMESPathCheck("properties.template.scale.rules[0].name", "http-scale-rule"),
            JMESPathCheck("properties.template.scale.rules[0].http.metadata.concurrentRequests", "50"),
            JMESPathCheck("properties.template.scale.rules[0].http.metadata.key", "value"),
            JMESPathCheck("properties.template.scale.rules[0].http.auth[0].triggerParameter", "trigger"),
            JMESPathCheck("properties.template.scale.rules[0].http.auth[0].secretRef", "secretref"),
        ])
        clean_up_test_file(containerapp_file_name)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import importlib
import unittest

from azure.cli.command_modules.acs._consts import DecoratorMode
from azure.cli.command_modules.acs.base_decorator import (
    BaseAKSContext,
    BaseAKSManagedClusterDecorator,
    BaseAKSModels,
    BaseAKSParamDict,
    validate_decorator_mode,
)
from azure.cli.command_modules.acs.tests.latest.mocks import MockCLI, MockClient, MockCmd
from azure.cli.core.azclierror import CLIInternalError
from azure.cli.core.profiles import ResourceType


class BaseDecoratorHelperFunctionsTestCase(unittest.TestCase):
    def test_validate_decorator_mode(self):
        self.assertEqual(validate_decorator_mode(DecoratorMode.CREATE), True)
        self.assertEqual(validate_decorator_mode(DecoratorMode.UPDATE), True)
        self.assertEqual(validate_decorator_mode(DecoratorMode), False)
        self.assertEqual(validate_decorator_mode(1), False)
        self.assertEqual(validate_decorator_mode("1"), False)
        self.assertEqual(validate_decorator_mode(True), False)
        self.assertEqual(validate_decorator_mode({}), False)


class BaseAKSModelsTestCase(unittest.TestCase):
    def setUp(self):
        self.cli_ctx = MockCLI()
        self.cmd = MockCmd(self.cli_ctx)

    def test_models(self):
        # load models directly (instead of through the `get_sdk` method provided by the cli component)
        module_name = "azure.mgmt.containerservice.models"
        module = importlib.import_module(module_name)

        models = BaseAKSModels(self.cmd, ResourceType.MGMT_CONTAINERSERVICE)
        self.assertEqual(models.model_module, module)
        self.assertEqual(models.AgentPool, module.AgentPool)


class BaseAKSParamDictTestCase(unittest.TestCase):
    def test__init__(self):
        # fail on not passing dictionary-like parameters
        with self.assertRaises(CLIInternalError):
            BaseAKSParamDict([])

    def test_get(self):
        param_dict = BaseAKSParamDict({"abc": "xyz"})
        self.assertEqual(param_dict.get("abc"), "xyz")

        param_dict_2 = BaseAKSParamDict({"a": None, "ab": False, "abc": ""})
        self.assertEqual(param_dict_2.get("a", True), True)
        self.assertEqual(param_dict_2.get("a", "xyz"), "xyz")
        self.assertEqual(param_dict_2.get("ab", True), False)
        self.assertEqual(param_dict_2.get("abc", True), "")
        self.assertEqual(param_dict_2.get("abcd", True), True)

    def test_keys(self):
        param_dict = BaseAKSParamDict({"abc": "xyz"})
        self.assertEqual(list(param_dict.keys()), ["abc"])

    def test_values(self):
        param_dict = BaseAKSParamDict({"abc": "xyz"})
        self.assertEqual(list(param_dict.values()), ["xyz"])

    def test_items(self):
        param_dict = BaseAKSParamDict({"abc": "xyz"})
        self.assertEqual(list(param_dict.items()), [("abc", "xyz")])

    def test_print_usage_statistics(self):
        param_dict = BaseAKSParamDict({"abc": "xyz", "def": 100})
        param_dict.print_usage_statistics()


class BaseAKSContextTestCase(unittest.TestCase):
    def setUp(self):
        self.cli_ctx = MockCLI()
        self.cmd = MockCmd(self.cli_ctx)
        self.models = BaseAKSModels(self.cmd, ResourceType.MGMT_CONTAINERSERVICE)

    def test__init__(self):
        # fail on not passing dictionary-like parameters
        with self.assertRaises(CLIInternalError):
            BaseAKSContext(self.cmd, [], self.models, decorator_mode=DecoratorMode.CREATE)
        # fail on not passing decorator_mode with Enum type DecoratorMode
        with self.assertRaises(CLIInternalError):
            BaseAKSContext(self.cmd, BaseAKSParamDict({}), self.models, decorator_mode=1)

    def test_get_intermediate(self):
        ctx_1 = BaseAKSContext(self.cmd, BaseAKSParamDict({}), self.models, decorator_mode=DecoratorMode.CREATE)
        self.assertEqual(
            ctx_1.get_intermediate("fake-intermediate", "not found"),
            "not found",
        )

    def test_set_intermediate(self):
        ctx_1 = BaseAKSContext(self.cmd, BaseAKSParamDict({}), self.models, decorator_mode=DecoratorMode.CREATE)
        ctx_1.set_intermediate("test-intermediate", "test-intermediate-value")
        self.assertEqual(
            ctx_1.get_intermediate("test-intermediate"),
            "test-intermediate-value",
        )
        ctx_1.set_intermediate("test-intermediate", "new-test-intermediate-value")
        self.assertEqual(
            ctx_1.get_intermediate("test-intermediate"),
            "test-intermediate-value",
        )
        ctx_1.set_intermediate(
            "test-intermediate",
            "new-test-intermediate-value",
            overwrite_exists=True,
        )
        self.assertEqual(
            ctx_1.get_intermediate("test-intermediate"),
            "new-test-intermediate-value",
        )

    def test_remove_intermediate(self):
        ctx_1 = BaseAKSContext(self.cmd, BaseAKSParamDict({}), self.models, decorator_mode=DecoratorMode.CREATE)
        ctx_1.set_intermediate("test-intermediate", "test-intermediate-value")
        self.assertEqual(
            ctx_1.get_intermediate("test-intermediate"),
            "test-intermediate-value",
        )
        ctx_1.remove_intermediate("test-intermediate")
        self.assertEqual(ctx_1.get_intermediate("test-intermediate"), None)


class BaseAKSManagedClusterDecoratorTestCase(unittest.TestCase):
    def setUp(self):
        self.cli_ctx = MockCLI()
        self.cmd = MockCmd(self.cli_ctx)
        self.client = MockClient()
        self.test_models = BaseAKSModels(self.cmd, ResourceType.MGMT_CONTAINERSERVICE)

    def test_init(self):
        dec_1 = BaseAKSManagedClusterDecorator(self.cmd, self.client)
        self.assertEqual(dec_1.cmd, self.cmd)
        self.assertEqual(dec_1.client, self.client)

    def test_init_models(self):
        dec_1 = BaseAKSManagedClusterDecorator(self.cmd, self.client)
        with self.assertRaises(NotImplementedError):
            dec_1.init_models()

    def test_init_context(self):
        dec_1 = BaseAKSManagedClusterDecorator(self.cmd, self.client)
        with self.assertRaises(NotImplementedError):
            dec_1.init_context()

    def test_check_is_postprocessing_required(self):
        dec_1 = BaseAKSManagedClusterDecorator(self.cmd, self.client)
        mc_1 = self.test_models.ManagedCluster(location="test_location")
        with self.assertRaises(NotImplementedError):
            dec_1.check_is_postprocessing_required(mc_1)

    def test_immediate_processing_after_request(self):
        dec_1 = BaseAKSManagedClusterDecorator(self.cmd, self.client)
        mc_1 = self.test_models.ManagedCluster(location="test_location")
        with self.assertRaises(NotImplementedError):
            dec_1.immediate_processing_after_request(mc_1)

    def test_postprocessing_after_mc_created(self):
        dec_1 = BaseAKSManagedClusterDecorator(self.cmd, self.client)
        mc_1 = self.test_models.ManagedCluster(location="test_location")
        with self.assertRaises(NotImplementedError):
            dec_1.postprocessing_after_mc_created(mc_1)

    def test_put_mc(self):
        dec_1 = BaseAKSManagedClusterDecorator(self.cmd, self.client)
        mc_1 = self.test_models.ManagedCluster(location="test_location")
        with self.assertRaises(NotImplementedError):
            dec_1.put_mc(mc_1)


if __name__ == "__main__":
    unittest.main()



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import os
from azure.cli.testsdk import (ScenarioTest)

TEST_DIR = os.path.abspath(os.path.join(os.path.abspath(__file__), '..'))
TEST_LOCATION = os.getenv("CLITestLocation") if os.getenv("CLITestLocation") else "eastus"
STAGE_LOCATION = "northcentralusstage"

def write_test_file(filename, content):
    test_file = open(filename, "w", encoding='utf-8')
    _ = test_file.write(content)
    test_file.close()


def clean_up_test_file(filename):
    if os.path.exists(filename):
        os.remove(filename)


class ContainerappComposePreviewScenarioTest(ScenarioTest):
    pass



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
#
# Code generated by aaz-dev-tools
# --------------------------------------------------------------------------------------------

# pylint: skip-file
# flake8: noqa

from azure.cli.core.aaz import *


@register_command(
    "network virtual-appliance inbound-security-rule create",
)
class Create(AAZCommand):
    """Create the specified Network Virtual Appliance Inbound Security Rules.

    :example: Create Inbound Security Rule of Permanent Rule Type
        az network virtual-appliance inbound-security-rule create  --network-virtual-appliance-name "MyName" -g "MyRG" --subscription {subID} --rule-type "Permanent" --name "PermanentRuleCollection" --rules "[{name:'inboundRule',protocol:'TCP',destination-port-ranges:['80-120'],applies-on:['slbIP'],source-address-prefix:'*'}]"

    :example: Create Inbound Security Rule of AutoExpire Rule Type
        az network virtual-appliance inbound-security-rule create  --network-virtual-appliance-name "MyName" -g "MyRG" --subscription {subID} --rule-type "AutoExpire" --name "TemporaryRuleCollection" --rules "[{name:'inboundRule',protocol:'TCP',destination-port-ranges:['80-120'],applies-on:['publicnicipconfig'],source-address-prefix:'20.0.0.0/32'}]"
    """

    _aaz_info = {
        "version": "2023-11-01",
        "resources": [
            ["mgmt-plane", "/subscriptions/{}/resourcegroups/{}/providers/microsoft.network/networkvirtualappliances/{}/inboundsecurityrules/{}", "2023-11-01"],
        ]
    }

    AZ_SUPPORT_NO_WAIT = True

    def _handler(self, command_args):
        super()._handler(command_args)
        return self.build_lro_poller(self._execute_operations, self._output)

    _args_schema = None

    @classmethod
    def _build_arguments_schema(cls, *args, **kwargs):
        if cls._args_schema is not None:
            return cls._args_schema
        cls._args_schema = super()._build_arguments_schema(*args, **kwargs)

        # define Arg Group ""

        _args_schema = cls._args_schema
        _args_schema.nva_name = AAZStrArg(
            options=["--nva-name"],
            help="The name of the Network Virtual Appliance.",
            required=True,
        )
        _args_schema.resource_group = AAZResourceGroupNameArg(
            required=True,
        )
        _args_schema.rule_collection_name = AAZStrArg(
            options=["-n", "--name", "--rule-collection-name"],
            help="The name of security rule collection.",
            required=True,
        )

        # define Arg Group "Parameters"

        _args_schema = cls._args_schema
        _args_schema.id = AAZStrArg(
            options=["--id"],
            arg_group="Parameters",
            help="Resource ID.",
        )

        # define Arg Group "Properties"

        _args_schema = cls._args_schema
        _args_schema.rule_type = AAZStrArg(
            options=["--rule-type"],
            arg_group="Properties",
            help="Rule Type. This should be either AutoExpire or Permanent. Auto Expire Rule only creates NSG rules. Permanent Rule creates NSG rule and SLB LB Rule.",
            enum={"AutoExpire": "AutoExpire", "Permanent": "Permanent"},
        )
        _args_schema.rules = AAZListArg(
            options=["--rules"],
            arg_group="Properties",
            help="List of allowed rules.",
        )

        rules = cls._args_schema.rules
        rules.Element = AAZObjectArg()

        _element = cls._args_schema.rules.Element
        _element.applies_on = AAZListArg(
            options=["applies-on"],
            help="Public IP name in case of Permanent Rule type & Interface Name in case of Auto Expire Rule type",
        )
        _element.destination_port_range = AAZIntArg(
            options=["destination-port-range"],
            help="NVA port ranges to be opened up. One needs to provide specific ports.",
            fmt=AAZIntArgFormat(
                maximum=65535,
                minimum=0,
            ),
        )
        _element.destination_port_ranges = AAZListArg(
            options=["destination-port-ranges"],
            help="NVA port ranges to be opened up. One can provide a range of ports. Allowed port value between 0 and 65535.",
        )
        _element.name = AAZStrArg(
            options=["name"],
            help="Name of the rule.",
        )
        _element.protocol = AAZStrArg(
            options=["protocol"],
            help="Protocol. This should be either TCP or UDP.",
            enum={"TCP": "TCP", "UDP": "UDP"},
        )
        _element.source_address_prefix = AAZStrArg(
            options=["source-address-prefix"],
            help="The CIDR or source IP range.",
        )

        applies_on = cls._args_schema.rules.Element.applies_on
        applies_on.Element = AAZStrArg()

        destination_port_ranges = cls._args_schema.rules.Element.destination_port_ranges
        destination_port_ranges.Element = AAZStrArg()
        return cls._args_schema

    def _execute_operations(self):
        self.pre_operations()
        yield self.InboundSecurityRuleCreateOrUpdate(ctx=self.ctx)()
        self.post_operations()

    @register_callback
    def pre_operations(self):
        pass

    @register_callback
    def post_operations(self):
        pass

    def _output(self, *args, **kwargs):
        result = self.deserialize_output(self.ctx.vars.instance, client_flatten=True)
        return result

    class InboundSecurityRuleCreateOrUpdate(AAZHttpOperation):
        CLIENT_TYPE = "MgmtClient"

        def __call__(self, *args, **kwargs):
            request = self.make_request()
            session = self.client.send_request(request=request, stream=False, **kwargs)
            if session.http_response.status_code in [202]:
                return self.client.build_lro_polling(
                    self.ctx.args.no_wait,
                    session,
                    self.on_200_201,
                    self.on_error,
                    lro_options={"final-state-via": "azure-async-operation"},
                    path_format_arguments=self.url_parameters,
                )
            if session.http_response.status_code in [200, 201]:
                return self.client.build_lro_polling(
                    self.ctx.args.no_wait,
                    session,
                    self.on_200_201,
                    self.on_error,
                    lro_options={"final-state-via": "azure-async-operation"},
                    path_format_arguments=self.url_parameters,
                )

            return self.on_error(session.http_response)

        @property
        def url(self):
            return self.client.format_url(
                "/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/Microsoft.Network/networkVirtualAppliances/{networkVirtualApplianceName}/inboundSecurityRules/{ruleCollectionName}",
                **self.url_parameters
            )

        @property
        def method(self):
            return "PUT"

        @property
        def error_format(self):
            return "ODataV4Format"

        @property
        def url_parameters(self):
            parameters = {
                **self.serialize_url_param(
                    "networkVirtualApplianceName", self.ctx.args.nva_name,
                    required=True,
                ),
                **self.serialize_url_param(
                    "resourceGroupName", self.ctx.args.resource_group,
                    required=True,
                ),
                **self.serialize_url_param(
                    "ruleCollectionName", self.ctx.args.rule_collection_name,
                    required=True,
                ),
                **self.serialize_url_param(
                    "subscriptionId", self.ctx.subscription_id,
                    required=True,
                ),
            }
            return parameters

        @property
        def query_parameters(self):
            parameters = {
                **self.serialize_query_param(
                    "api-version", "2023-11-01",
                    required=True,
                ),
            }
            return parameters

        @property
        def header_parameters(self):
            parameters = {
                **self.serialize_header_param(
                    "Content-Type", "application/json",
                ),
                **self.serialize_header_param(
                    "Accept", "application/json",
                ),
            }
            return parameters

        @property
        def content(self):
            _content_value, _builder = self.new_content_builder(
                self.ctx.args,
                typ=AAZObjectType,
                typ_kwargs={"flags": {"required": True, "client_flatten": True}}
            )
            _builder.set_prop("id", AAZStrType, ".id")
            _builder.set_prop("name", AAZStrType, ".rule_collection_name")
            _builder.set_prop("properties", AAZObjectType, typ_kwargs={"flags": {"client_flatten": True}})

            properties = _builder.get(".properties")
            if properties is not None:
                properties.set_prop("ruleType", AAZStrType, ".rule_type")
                properties.set_prop("rules", AAZListType, ".rules")

            rules = _builder.get(".properties.rules")
            if rules is not None:
                rules.set_elements(AAZObjectType, ".")

            _elements = _builder.get(".properties.rules[]")
            if _elements is not None:
                _elements.set_prop("appliesOn", AAZListType, ".applies_on")
                _elements.set_prop("destinationPortRange", AAZIntType, ".destination_port_range")
                _elements.set_prop("destinationPortRanges", AAZListType, ".destination_port_ranges")
                _elements.set_prop("name", AAZStrType, ".name")
                _elements.set_prop("protocol", AAZStrType, ".protocol")
                _elements.set_prop("sourceAddressPrefix", AAZStrType, ".source_address_prefix")

            applies_on = _builder.get(".properties.rules[].appliesOn")
            if applies_on is not None:
                applies_on.set_elements(AAZStrType, ".")

            destination_port_ranges = _builder.get(".properties.rules[].destinationPortRanges")
            if destination_port_ranges is not None:
                destination_port_ranges.set_elements(AAZStrType, ".")

            return self.serialize_content(_content_value)

        def on_200_201(self, session):
            data = self.deserialize_http_content(session)
            self.ctx.set_var(
                "instance",
                data,
                schema_builder=self._build_schema_on_200_201
            )

        _schema_on_200_201 = None

        @classmethod
        def _build_schema_on_200_201(cls):
            if cls._schema_on_200_201 is not None:
                return cls._schema_on_200_201

            cls._schema_on_200_201 = AAZObjectType()

            _schema_on_200_201 = cls._schema_on_200_201
            _schema_on_200_201.etag = AAZStrType(
                flags={"read_only": True},
            )
            _schema_on_200_201.id = AAZStrType()
            _schema_on_200_201.name = AAZStrType()
            _schema_on_200_201.properties = AAZObjectType(
                flags={"client_flatten": True},
            )
            _schema_on_200_201.type = AAZStrType(
                flags={"read_only": True},
            )

            properties = cls._schema_on_200_201.properties
            properties.provisioning_state = AAZStrType(
                serialized_name="provisioningState",
                flags={"read_only": True},
            )
            properties.rule_type = AAZStrType(
                serialized_name="ruleType",
            )
            properties.rules = AAZListType()

            rules = cls._schema_on_200_201.properties.rules
            rules.Element = AAZObjectType()

            _element = cls._schema_on_200_201.properties.rules.Element
            _element.applies_on = AAZListType(
                serialized_name="appliesOn",
            )
            _element.destination_port_range = AAZIntType(
                serialized_name="destinationPortRange",
            )
            _element.destination_port_ranges = AAZListType(
                serialized_name="destinationPortRanges",
            )
            _element.name = AAZStrType()
            _element.protocol = AAZStrType()
            _element.source_address_prefix = AAZStrType(
                serialized_name="sourceAddressPrefix",
            )

            applies_on = cls._schema_on_200_201.properties.rules.Element.applies_on
            applies_on.Element = AAZStrType()

            destination_port_ranges = cls._schema_on_200_201.properties.rules.Element.destination_port_ranges
            destination_port_ranges.Element = AAZStrType()

            return cls._schema_on_200_201


class _CreateHelper:
    """Helper class for Create"""


__all__ = ["Create"]



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
import unittest
from unittest.mock import Mock, patch
from azure.cli.command_modules.acs.azuremonitormetrics.amw.helper import get_amw_region
from azure.cli.command_modules.acs._client_factory import get_resources_client
from azure.core.exceptions import HttpResponseError

class TestAMWRegion(unittest.TestCase):
    @patch('azure.cli.command_modules.acs.azuremonitormetrics.amw.helper.get_resources_client')
    def test_get_amw_region(self, mock_get_resources_client):
        cmd = Mock()
        azure_monitor_workspace_resource_id = "/subscriptions/test_subscription/resourceGroups/test_resource_group/providers/Microsoft.OperationalInsights/workspaces/test_workspace"
        
        mock_resources_client = Mock()
        mock_get_resources_client.return_value = mock_resources_client
        mock_resource = Mock()
        mock_resource.location = 'East US'
        mock_resources_client.get_by_id.return_value = mock_resource

        result = get_amw_region(cmd, azure_monitor_workspace_resource_id)
        self.assertEqual(result, 'eastus')
        
        mock_resource.location = 'Central US (EUAP)'
        result = get_amw_region(cmd, azure_monitor_workspace_resource_id)
        self.assertEqual(result, 'centraluseuap')
        
        mock_resource.location = 'West US 2'
        result = get_amw_region(cmd, azure_monitor_workspace_resource_id)
        self.assertEqual(result, 'westus2')

    @patch('azure.cli.command_modules.acs.azuremonitormetrics.amw.helper.get_resources_client')
    def test_get_amw_region_with_invalid_characters(self, mock_get_resources_client):
        cmd = Mock()
        azure_monitor_workspace_resource_id = "/subscriptions/test_subscription/resourceGroups/test_resource_group/providers/Microsoft.OperationalInsights/workspaces/test_workspace"
        
        mock_resources_client = Mock()
        mock_get_resources_client.return_value = mock_resources_client
        mock_resource = Mock()
        mock_resource.location = 'East-US@123'
        mock_resources_client.get_by_id.return_value = mock_resource

        result = get_amw_region(cmd, azure_monitor_workspace_resource_id)
        self.assertEqual(result, 'eastus123')

    @patch('azure.cli.command_modules.acs.azuremonitormetrics.amw.helper.get_resources_client')
    def test_get_amw_region_http_error(self, mock_get_resources_client):
        cmd = Mock()
        azure_monitor_workspace_resource_id = "/subscriptions/test_subscription/resourceGroups/test_resource_group/providers/Microsoft.OperationalInsights/workspaces/test_workspace"
        
        mock_resources_client = Mock()
        mock_get_resources_client.return_value = mock_resources_client
        mock_resources_client.get_by_id.side_effect = HttpResponseError("An error occurred")

        with self.assertRaises(HttpResponseError):
            get_amw_region(cmd, azure_monitor_workspace_resource_id)

if __name__ == '__main__':
    unittest.main()



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from typing import Tuple

from azure.cli.command_modules.acs.azurecontainerstorage._consts import (
    CONST_ACSTOR_ALL,
    CONST_ACSTOR_EXT_INSTALLATION_NAME,
    CONST_ACSTOR_IO_ENGINE_LABEL_KEY,
    CONST_ACSTOR_K8S_EXTENSION_NAME,
    CONST_ACSTOR_V1_K8S_EXTENSION_NAME,
    CONST_DISK_TYPE_EPHEMERAL_VOLUME_ONLY,
    CONST_DISK_TYPE_PV_WITH_ANNOTATION,
    CONST_EPHEMERAL_NVME_PERF_TIER_BASIC,
    CONST_EPHEMERAL_NVME_PERF_TIER_PREMIUM,
    CONST_EPHEMERAL_NVME_PERF_TIER_STANDARD,
    CONST_ACSTOR_V1_EXT_INSTALLATION_NAME,
    CONST_K8S_EXTENSION_CLIENT_FACTORY_MOD_NAME,
    CONST_K8S_EXTENSION_CUSTOM_MOD_NAME,
    CONST_K8S_EXTENSION_NAME,
    CONST_STORAGE_POOL_OPTION_NVME,
    CONST_STORAGE_POOL_OPTION_SSD,
    CONST_STORAGE_POOL_TYPE_AZURE_DISK,
    CONST_STORAGE_POOL_TYPE_ELASTIC_SAN,
    CONST_STORAGE_POOL_TYPE_EPHEMERAL_DISK,
)
from azure.cli.command_modules.acs._roleassignments import (
    add_role_assignment,
    build_role_scope,
    delete_role_assignments,
)
from azure.cli.core.azclierror import ResourceNotFoundError, UnknownError
from knack.log import get_logger

logger = get_logger(__name__)

# Global cache to store VM SKU information.
vm_sku_details_cache = {}


def validate_storagepool_creation(
    cmd,
    subscription_id,
    node_resource_group,
    kubelet_identity_object_id,
    storage_pool_type,
):
    if storage_pool_type == CONST_STORAGE_POOL_TYPE_ELASTIC_SAN:
        role_assignment_success = perform_role_operations_on_managed_rg(
            cmd,
            subscription_id,
            node_resource_group,
            kubelet_identity_object_id,
            True
        )

        if not role_assignment_success:
            raise UnknownError(
                f"Cannot set --enable-azure-container-storage to {CONST_STORAGE_POOL_TYPE_ELASTIC_SAN}. "
                "Unable to add Role Assignments needed for Elastic SAN storage pools to be functional. "
                "Please check with your admin for permissions."
            )


def perform_role_operations_on_managed_rg(
    cmd,
    subscription_id,
    node_resource_group,
    kubelet_identity_object_id,
    assign
):
    managed_rg_role_scope = build_role_scope(node_resource_group, None, subscription_id)
    roles = ["Azure Container Storage Operator"]
    result = True

    for role in roles:
        try:
            if assign:
                result = add_role_assignment(
                    cmd,
                    role,
                    kubelet_identity_object_id,
                    scope=managed_rg_role_scope,
                    delay=0,
                )
            else:
                # NOTE: delete_role_assignments accepts cli_ctx
                # instead of cmd unlike add_role_assignment.
                result = delete_role_assignments(
                    cmd.cli_ctx,
                    role,
                    kubelet_identity_object_id,
                    scope=managed_rg_role_scope,
                    delay=0,
                )

            if not result:
                break
        except Exception:  # pylint: disable=broad-except
            break
    else:
        return True

    if not assign:
        logger.error("\nUnable to revoke Role Assignments if any, added for Azure Container Storage.")

    return False


def get_k8s_extension_module(module_name):
    try:
        # adding the installed extension in the path
        from azure.cli.core.extension.operations import add_extension_to_path
        add_extension_to_path(CONST_K8S_EXTENSION_NAME)
        # import the extension module
        from importlib import import_module
        azext_custom = import_module(module_name)
        return azext_custom
    except ImportError:
        raise UnknownError(  # pylint: disable=raise-missing-from
            "Please add CLI extension `k8s-extension` for performing Azure Container Storage operations.\n"
            "Run command `az extension add --name k8s-extension`"
        )


def check_if_extension_is_installed(cmd, resource_group, cluster_name) -> bool:
    client_factory = get_k8s_extension_module(CONST_K8S_EXTENSION_CLIENT_FACTORY_MOD_NAME)
    client = client_factory.cf_k8s_extension_operation(cmd.cli_ctx)
    k8s_extension_custom_mod = get_k8s_extension_module(CONST_K8S_EXTENSION_CUSTOM_MOD_NAME)
    return_val = True
    try:
        extension = k8s_extension_custom_mod.show_k8s_extension(
            client,
            resource_group,
            cluster_name,
            CONST_ACSTOR_V1_EXT_INSTALLATION_NAME,
            "managedClusters",
        )

        extension_type = extension.extension_type.lower()
        if extension_type != CONST_ACSTOR_V1_K8S_EXTENSION_NAME:
            return_val = False
    except:  # pylint: disable=bare-except
        return_val = False

    return return_val


def get_extension_installed_and_cluster_configs_v1(
    cmd,
    resource_group,
    cluster_name,
    agentpool_profiles
) -> Tuple[bool, bool, bool, bool, bool, float, str, str]:
    client_factory = get_k8s_extension_module(CONST_K8S_EXTENSION_CLIENT_FACTORY_MOD_NAME)
    client = client_factory.cf_k8s_extension_operation(cmd.cli_ctx)
    k8s_extension_custom_mod = get_k8s_extension_module(CONST_K8S_EXTENSION_CUSTOM_MOD_NAME)
    is_extension_installed = False
    is_elasticSan_enabled = False
    is_azureDisk_enabled = False
    is_ephemeralDisk_nvme_enabled = False
    is_ephemeralDisk_localssd_enabled = False
    ephemeral_disk_volume_type = CONST_DISK_TYPE_EPHEMERAL_VOLUME_ONLY
    perf_tier = CONST_EPHEMERAL_NVME_PERF_TIER_STANDARD
    resource_cpu_value = -1

    try:
        extension = k8s_extension_custom_mod.show_k8s_extension(
            client,
            resource_group,
            cluster_name,
            CONST_ACSTOR_V1_EXT_INSTALLATION_NAME,
            "managedClusters",
        )

        extension_type = extension.extension_type.lower()
        is_extension_installed = extension_type == CONST_ACSTOR_V1_K8S_EXTENSION_NAME
        config_settings = extension.configuration_settings

        if is_extension_installed and config_settings is not None:
            is_cli_operation_active = config_settings.get("global.cli.activeControl", "False") == "True"
            if is_cli_operation_active:
                is_azureDisk_enabled = (
                    config_settings.get("global.cli.storagePool.azureDisk.enabled", "False") == "True"
                )
                is_elasticSan_enabled = (
                    config_settings.get("global.cli.storagePool.elasticSan.enabled", "False") == "True"
                )
                is_ephemeralDisk_nvme_enabled = (
                    config_settings.get("global.cli.storagePool.ephemeralDisk.nvme.enabled", "False") == "True"
                )
                is_ephemeralDisk_localssd_enabled = (
                    config_settings.get("global.cli.storagePool.ephemeralDisk.temp.enabled", "False") == "True"
                )
                cpu_value = config_settings.get("global.cli.resources.ioEngine.cpu", "1")
                enable_ephemeral_bypass_annotation = (
                    config_settings.get(
                        "global.cli.storagePool.ephemeralDisk.enableEphemeralBypassAnnotation", "False"
                    ) == "True"
                )
                perf_tier = config_settings.get(
                    "global.cli.storagePool.ephemeralDisk.nvme.perfTier",
                    CONST_EPHEMERAL_NVME_PERF_TIER_STANDARD
                )

                if perf_tier.lower() == CONST_EPHEMERAL_NVME_PERF_TIER_BASIC.lower():
                    perf_tier = CONST_EPHEMERAL_NVME_PERF_TIER_BASIC
                elif perf_tier.lower() == CONST_EPHEMERAL_NVME_PERF_TIER_PREMIUM.lower():
                    perf_tier = CONST_EPHEMERAL_NVME_PERF_TIER_PREMIUM
                else:
                    perf_tier = CONST_EPHEMERAL_NVME_PERF_TIER_STANDARD

                if enable_ephemeral_bypass_annotation:
                    ephemeral_disk_volume_type = CONST_DISK_TYPE_PV_WITH_ANNOTATION
                resource_cpu_value = float(cpu_value)
            else:
                # For versions where "global.cli.activeControl" were not set it signifies
                # that selective control of storgepool type was not yet defined.
                # Hence, all the storagepool types are active and io engine core count is 1.
                is_azureDisk_enabled = is_elasticSan_enabled = is_ephemeralDisk_localssd_enabled = True
                resource_cpu_value = 1

                # Determine if ephemeral NVMe was active based on the labelled nodepools present in cluster.
                for agentpool in agentpool_profiles:
                    vm_size = agentpool.vm_size
                    node_labels = agentpool.node_labels
                    if (node_labels is not None and
                            node_labels.get(CONST_ACSTOR_IO_ENGINE_LABEL_KEY) is not None and
                            vm_size.lower().startswith('standard_l')):
                        is_ephemeralDisk_nvme_enabled = True
                        break

    except:  # pylint: disable=bare-except
        is_extension_installed = False

    return (
        is_extension_installed,
        is_azureDisk_enabled,
        is_elasticSan_enabled,
        is_ephemeralDisk_localssd_enabled,
        is_ephemeralDisk_nvme_enabled,
        resource_cpu_value,
        ephemeral_disk_volume_type,
        perf_tier
    )


def get_extension_installed_and_cluster_configs(
    cmd,
    resource_group,
    cluster_name
) -> Tuple[bool, bool, bool]:
    client_factory = get_k8s_extension_module(CONST_K8S_EXTENSION_CLIENT_FACTORY_MOD_NAME)
    client = client_factory.cf_k8s_extension_operation(cmd.cli_ctx)
    k8s_extension_custom_mod = get_k8s_extension_module(CONST_K8S_EXTENSION_CUSTOM_MOD_NAME)
    is_extension_installed = False
    is_ephemeral_disk_enabled = False
    is_elastic_san_enabled = False

    try:
        extension = k8s_extension_custom_mod.show_k8s_extension(
            client,
            resource_group,
            cluster_name,
            CONST_ACSTOR_EXT_INSTALLATION_NAME,
            "managedClusters",
        )

        extension_type = extension.extension_type.lower()
        is_extension_installed = extension_type == CONST_ACSTOR_K8S_EXTENSION_NAME
        config_settings = extension.configuration_settings

        if is_extension_installed and config_settings is not None:
            is_ephemeral_disk_enabled = (
                config_settings.get("csiDriverConfigs.local-csi-driver.enabled", "False") == "True"
            )
            is_elastic_san_enabled = (
                config_settings.get("csiDriverConfigs.azuresan-csi-driver.enabled", "False") == "True"
            )

    except:  # pylint: disable=bare-except
        is_extension_installed = False

    return (
        is_extension_installed,
        is_ephemeral_disk_enabled,
        is_elastic_san_enabled,
    )


def should_delete_extension(storage_options_to_remove) -> bool:
    return (
        storage_options_to_remove in [True, CONST_ACSTOR_ALL] or
        (isinstance(storage_options_to_remove, list) and CONST_ACSTOR_ALL in storage_options_to_remove)
    )


def get_container_storage_extension_installed(
    cmd,
    resource_group,
    cluster_name,
    extension_name,
) -> Tuple[bool, str]:

    client_factory = get_k8s_extension_module(CONST_K8S_EXTENSION_CLIENT_FACTORY_MOD_NAME)
    client = client_factory.cf_k8s_extension_operation(cmd.cli_ctx)
    k8s_extension_custom_mod = get_k8s_extension_module(CONST_K8S_EXTENSION_CUSTOM_MOD_NAME)
    is_extension_installed = False
    extension_version = ""

    try:
        extension = k8s_extension_custom_mod.show_k8s_extension(
            client,
            resource_group,
            cluster_name,
            extension_name,
            "managedClusters",
        )
        is_extension_installed = True
        extension_version = extension.current_version
    except ResourceNotFoundError:
        # Extension not found, which is expected if not installed.
        is_extension_installed = False
    return is_extension_installed, extension_version


def get_initial_resource_value_args(
    storage_pool_type,
    storage_pool_option,
    nodepool_skus,
    ephemeral_nvme_perf_tier,
):
    core_value = memory_value = hugepages_value = hugepages_number = 0
    if (storage_pool_type == CONST_STORAGE_POOL_TYPE_AZURE_DISK or
        (storage_pool_type == CONST_STORAGE_POOL_TYPE_EPHEMERAL_DISK and
         storage_pool_option == CONST_STORAGE_POOL_OPTION_SSD)):
        core_value = 1
        memory_value = 1
        hugepages_value = 1
        hugepages_number = 512
    elif (storage_pool_type == CONST_STORAGE_POOL_TYPE_EPHEMERAL_DISK and
          storage_pool_option == CONST_STORAGE_POOL_OPTION_NVME):
        core_value = _get_ephemeral_nvme_cpu_value_based_on_vm_size_perf_tier(nodepool_skus, ephemeral_nvme_perf_tier)
        memory_value = 2
        hugepages_value = 2
        hugepages_number = 1024

    return _generate_k8s_extension_resource_args(
        core_value,
        memory_value,
        hugepages_value,
        hugepages_number,
    )


def get_current_resource_value_args(
    is_azureDisk_enabled,
    is_elasticSan_enabled,
    is_ephemeralDisk_localssd_enabled,
    is_ephemeralDisk_nvme_enabled,
    ephemeral_nvme_perf_tier,
    current_core_value=None,
    nodepool_skus=None,
):
    (
        current_core_value,
        current_memory_value,
        current_hugepages_value,
        current_hugepages_number,
    ) = _get_current_resource_values(
        is_azureDisk_enabled,
        is_elasticSan_enabled,
        is_ephemeralDisk_localssd_enabled,
        is_ephemeralDisk_nvme_enabled,
        ephemeral_nvme_perf_tier,
        current_core_value,
        nodepool_skus,
    )

    return _generate_k8s_extension_resource_args(
        current_core_value,
        current_memory_value,
        current_hugepages_value,
        current_hugepages_number,
    )


def get_desired_resource_value_args(
    storage_pool_type,
    storage_pool_option,
    ephemeral_nvme_perf_tier,
    current_core_value,
    is_azureDisk_enabled,
    is_elasticSan_enabled,
    is_ephemeralDisk_localssd_enabled,
    is_ephemeralDisk_nvme_enabled,
    perf_tier_updated,
    nodepool_skus,
    is_enabling_op,
):
    (
        current_core_value,
        current_memory_value,
        current_hugepages_value,
        current_hugepages_number,
    ) = _get_current_resource_values(
        is_azureDisk_enabled,
        is_elasticSan_enabled,
        is_ephemeralDisk_localssd_enabled,
        is_ephemeralDisk_nvme_enabled,
        ephemeral_nvme_perf_tier,
        current_core_value,
        nodepool_skus,
    )

    updated_core_value = updated_memory_value = \
        updated_hugepages_value = updated_hugepages_number = 0

    if is_enabling_op:
        if storage_pool_type == CONST_STORAGE_POOL_TYPE_AZURE_DISK or \
           (storage_pool_type == CONST_STORAGE_POOL_TYPE_EPHEMERAL_DISK and
                storage_pool_option == CONST_STORAGE_POOL_OPTION_SSD):
            updated_core_value = 1
            updated_memory_value = 1
            updated_hugepages_value = 1
            updated_hugepages_number = 512
        elif (storage_pool_type == CONST_STORAGE_POOL_TYPE_EPHEMERAL_DISK and
              (storage_pool_option == CONST_STORAGE_POOL_OPTION_NVME or
               is_ephemeralDisk_nvme_enabled or perf_tier_updated)):
            updated_core_value = _get_ephemeral_nvme_cpu_value_based_on_vm_size_perf_tier(
                nodepool_skus,
                ephemeral_nvme_perf_tier,
            )
            updated_memory_value = 2
            updated_hugepages_value = 2
            updated_hugepages_number = 1024

        if not perf_tier_updated:
            # For an operation where we are not modifying the perf tiers for nvme,
            # we have decided the updated value based on the type we are enabling.
            # Now, we compare and check if the current values are already greater
            # than that and if so, we preserve the current values.
            updated_core_value = current_core_value \
                if current_core_value > updated_core_value else updated_core_value
        updated_memory_value = current_memory_value \
            if current_memory_value > updated_memory_value else updated_memory_value
        updated_hugepages_value = current_hugepages_value \
            if current_hugepages_value > updated_hugepages_value else updated_hugepages_value
        updated_hugepages_number = current_hugepages_number \
            if current_hugepages_number > updated_hugepages_number else updated_hugepages_number
    else:
        # If we are disabling AzureDisk storagepool but EphemeralDisk(any) is
        # still enabled, or if we are disabling Ephemeral LocalSSD but
        # AzureDisk or Ephemeral NVMe storagepool is still enabled, or
        # if we are disabling ElasticSan storagepool but AzureDisk or any
        # EphemeralDisk storagepool type is still enabled,
        # then we will preserve the current resource values.
        is_disabled_type_smaller_than_active_types = (
            (storage_pool_type == CONST_STORAGE_POOL_TYPE_ELASTIC_SAN and
                (is_azureDisk_enabled or is_ephemeralDisk_nvme_enabled or is_ephemeralDisk_localssd_enabled)) or
            (storage_pool_type == CONST_STORAGE_POOL_TYPE_AZURE_DISK and
                (is_ephemeralDisk_nvme_enabled or is_ephemeralDisk_localssd_enabled)) or
            (storage_pool_type == CONST_STORAGE_POOL_TYPE_EPHEMERAL_DISK and
                storage_pool_option == CONST_STORAGE_POOL_OPTION_SSD and
                (is_azureDisk_enabled or is_ephemeralDisk_nvme_enabled))
        )
        is_ephemeral_nvme_disabled_azureDisk_active = (
            storage_pool_type == CONST_STORAGE_POOL_TYPE_EPHEMERAL_DISK and
            ((storage_pool_option == CONST_STORAGE_POOL_OPTION_NVME and
                (is_ephemeralDisk_localssd_enabled or is_azureDisk_enabled)) or
                (storage_pool_option == CONST_ACSTOR_ALL and is_azureDisk_enabled))
        )
        if is_disabled_type_smaller_than_active_types:
            updated_core_value = current_core_value
            updated_memory_value = current_memory_value
            updated_hugepages_value = current_hugepages_value
            updated_hugepages_number = current_hugepages_number
        elif is_ephemeral_nvme_disabled_azureDisk_active:
            # If we are disabling Ephemeral NVMe storagepool but azureDisk is
            # still enabled, we will set the azureDisk storagepool type values.
            updated_core_value = 1
            updated_memory_value = 1
            updated_hugepages_value = 1
            updated_hugepages_number = 512

    return _generate_k8s_extension_resource_args(
        updated_core_value,
        updated_memory_value,
        updated_hugepages_value,
        updated_hugepages_number,
    )


def check_if_new_storagepool_creation_required(
    storage_pool_type,
    ephemeral_disk_volume_type,
    ephemeral_disk_nvme_perf_tier,
    existing_ephemeral_disk_volume_type,
    existing_ephemeral_nvme_perf_tier,
    is_extension_installed,
    is_ephemeralDisk_nvme_enabled,
    is_ephemeralDisk_localssd_enabled,
) -> bool:
    should_create_storagepool = (
        not is_extension_installed or
        not (is_ephemeralDisk_localssd_enabled or is_ephemeralDisk_nvme_enabled) or
        storage_pool_type != CONST_STORAGE_POOL_TYPE_EPHEMERAL_DISK or
        (
            (ephemeral_disk_volume_type is None or
                (existing_ephemeral_disk_volume_type.lower() == ephemeral_disk_volume_type.lower())) and
            (ephemeral_disk_nvme_perf_tier is None or
                (existing_ephemeral_nvme_perf_tier.lower() == ephemeral_disk_nvme_perf_tier.lower()))
        )
    )

    return should_create_storagepool


def generate_vm_sku_cache_for_region(cli_ctx, location=None):
    result = _get_vm_sku_details(cli_ctx, location)
    for vm_data in result:
        sku_name = vm_data.name.lower()
        capabilities = vm_data.capabilities
        cpu_value = -1
        nvme_enabled = False
        for entry in capabilities:
            if entry.name == 'vCPUs' and cpu_value == -1:
                cpu_value = int(entry.value)

            if entry.name == 'vCPUsAvailable':
                cpu_value = int(entry.value)

            if entry.name == 'NvmeDiskSizeInMiB':
                nvme_enabled = True

            vm_sku_details_cache[sku_name] = (cpu_value, nvme_enabled)


def get_vm_sku_details(sku_name):
    result = vm_sku_details_cache.get(sku_name)
    if result is None:
        return None, None
    return result


def _get_ephemeral_nvme_cpu_value_based_on_vm_size_perf_tier(nodepool_skus, perf_tier):
    cpu_value = -1
    multiplication_factor = 0.25
    if perf_tier.lower() == CONST_EPHEMERAL_NVME_PERF_TIER_BASIC.lower():
        multiplication_factor = 0.15
    elif perf_tier.lower() == CONST_EPHEMERAL_NVME_PERF_TIER_PREMIUM.lower():
        multiplication_factor = 0.5
    for vm_size in nodepool_skus:
        number_of_cores, _ = get_vm_sku_details(vm_size.lower())
        if number_of_cores is None:
            raise UnknownError(
                f"Unable to find details for virtual machine size {vm_size}."
            )
        if number_of_cores != -1:
            if cpu_value == -1:
                cpu_value = number_of_cores * multiplication_factor
            else:
                cpu_value = (number_of_cores * multiplication_factor) if \
                    (cpu_value > number_of_cores * multiplication_factor) else \
                    cpu_value
        else:
            raise UnknownError(
                f"Unable to determine the number of cores in nodepool of node size: {vm_size}"
            )

    # In any case when cpu_value = -1 or is lesser than 1,
    # set the value to 1.
    cpu_value = max(cpu_value, 1)
    return cpu_value


def _get_current_resource_values(
    is_azureDisk_enabled,
    is_elasticSan_enabled,
    is_ephemeralDisk_localssd_enabled,
    is_ephemeralDisk_nvme_enabled,
    ephemeral_nvme_perf_tier,
    current_core_value=None,
    nodepool_skus=None,
):
    # Setting these to default values set in the cluster when
    # all the storagepools used to be enabled by default.
    core_value = current_memory_value = current_hugepages_value = 1
    current_hugepages_number = 1024
    if is_elasticSan_enabled:
        core_value = 0
        current_memory_value = 0
        current_hugepages_value = 0
        current_hugepages_number = 0
    if is_azureDisk_enabled or is_ephemeralDisk_localssd_enabled:
        core_value = 1
        current_memory_value = 1
        current_hugepages_value = 1
        current_hugepages_number = 512
    if is_ephemeralDisk_nvme_enabled:
        if current_core_value is None and nodepool_skus is not None:
            core_value = _get_ephemeral_nvme_cpu_value_based_on_vm_size_perf_tier(
                nodepool_skus,
                ephemeral_nvme_perf_tier,
            )
        current_memory_value = 2
        current_hugepages_value = 2
        current_hugepages_number = 1024

    current_core_value = current_core_value if current_core_value is not None else core_value

    return (
        current_core_value,
        current_memory_value,
        current_hugepages_value,
        current_hugepages_number
    )


def _generate_k8s_extension_resource_args(
    core_value,
    memory_value,
    hugepages_value,
    hugepages_number,
):
    memory_value_str = str(memory_value) + "Gi"
    hugepages_value_str = str(hugepages_value) + "Gi"

    resource_args = [
        {"global.cli.resources.num_hugepages": hugepages_number},
        {"global.cli.resources.ioEngine.cpu": str(core_value)},
        {"global.cli.resources.ioEngine.memory": memory_value_str},
        {"global.cli.resources.ioEngine.hugepages2Mi": hugepages_value_str},
    ]

    return resource_args


def _get_vm_sku_details(cli_ctx, location=None):
    def _is_vm_in_required_location(desired_location, location_list):
        for val in location_list:
            if desired_location.lower() == val.lower():
                return True
        return False

    from azure.cli.command_modules.acs._client_factory import get_compute_client
    result = get_compute_client(cli_ctx).resource_skus.list()
    result = [x for x in result if x.resource_type.lower() == 'virtualmachines']
    if location:
        result = [r for r in result if _is_vm_in_required_location(location, r.locations)]
    return result



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import time
from azure.cli.testsdk.scenario_tests import AllowLargeResponse
from azure.cli.testsdk import (
    JMESPathCheck,
    NoneCheck,
    ResourceGroupPreparer,
    ScenarioTest)
from .constants import DEFAULT_LOCATION, SERVER_NAME_PREFIX, SERVER_NAME_MAX_LENGTH


class PostgreSQLFlexibleServerVnetMgmtScenarioTest(ScenarioTest):

    postgres_location = DEFAULT_LOCATION

    @AllowLargeResponse()
    @ResourceGroupPreparer(location=postgres_location)
    def test_postgres_flexible_server_vnet_mgmt_subnetid_and_privatednszoneid(self, resource_group):
        self._test_flexible_server_vnet_mgmt_subnetid_and_privatednszoneid(resource_group)

    @AllowLargeResponse()
    @ResourceGroupPreparer(location=postgres_location)
    def test_postgres_flexible_server_vnet_mgmt_vnetname_subnetname_and_privatednszoneid(self, resource_group):
        self._test_flexible_server_vnet_mgmt_vnetname_subnetname_and_privatednszoneid(resource_group)

    @AllowLargeResponse()
    @ResourceGroupPreparer(location=postgres_location, parameter_name='resource_group_vnet')
    @ResourceGroupPreparer(location=postgres_location, parameter_name='resource_group_private_dns_zone')
    @ResourceGroupPreparer(location=postgres_location, parameter_name='resource_group_server')
    def test_flexible_server_vnet_mgmt_subnetid_and_privatednszoneid_in_different_resource_groups(self, resource_group_vnet, resource_group_private_dns_zone, resource_group_server):
        self._test_flexible_server_vnet_mgmt_subnetid_and_privatednszoneid_in_different_resource_groups(resource_group_vnet, resource_group_private_dns_zone, resource_group_server)

    def _test_flexible_server_vnet_mgmt_subnetid_and_privatednszoneid(self, resource_group):
        if self.cli_ctx.local_context.is_on:
            self.cmd('config param-persist off')

        location = self.postgres_location
        server_name = self.create_random_name(SERVER_NAME_PREFIX, 32)
        server_vnet_name = self.create_random_name('VNET', SERVER_NAME_MAX_LENGTH)
        server_subnet_name = self.create_random_name('SUBNET', SERVER_NAME_MAX_LENGTH)
        server_vnet_prefixes = '10.0.0.0/16'
        server_subnet_prefixes = '10.0.0.0/24'
        server_private_dns_zone_name = '{}.private.postgres.database.azure.com'.format(server_name)

        # Create a virtual network and subnet for server.
        result_vnet = self.cmd('network vnet create -g {} -l {} -n {} --address-prefixes {} --subnet-name {} --subnet-prefixes {}'.format(
                 resource_group, location, server_vnet_name, server_vnet_prefixes, server_subnet_name, server_subnet_prefixes)).get_output_in_json()
        
        # Delegate the subnet to Microsoft.DBforPostgreSQL/flexibleServers.
        self.cmd('network vnet subnet update -g {} --vnet-name {} -n {} --delegations Microsoft.DBforPostgreSQL/flexibleServers'.format(
                 resource_group, server_vnet_name, server_subnet_name))
        
        # Create a private DNS zone for server.
        result_dns = self.cmd('network private-dns zone create -g {} -n {}'.format(resource_group, server_private_dns_zone_name)).get_output_in_json()

        # Scenario: Provision a server with supplied subnet identifier and private DNS zone identifier.
        self.cmd('postgres flexible-server create -g {} -n {} --subnet {} -l {} --private-dns-zone {} --yes'
                 .format(resource_group, server_name, result_vnet['newVNet']['subnets'][0]['id'], location, result_dns['id']))
        
        # Validate that the server is provisioned with the correct network configuration.
        self.cmd('postgres flexible-server show -g {} -n {}'
                 .format(resource_group, server_name),
                 checks=[
                     JMESPathCheck('network.privateDnsZoneArmResourceId', result_dns['id']),
                     JMESPathCheck('network.delegatedSubnetResourceId', result_vnet['newVNet']['subnets'][0]['id'])])

        # Scenario: Migrate network of the server from integrated in customer-managed network to integrated in Microsoft-managed network.
        self.cmd('postgres flexible-server migrate-network -g {} -n {}'.format(resource_group, server_name))
        
        self.cmd('postgres flexible-server show -g {} -n {}'
                    .format(resource_group, server_name),
                    checks=[
                        JMESPathCheck('network.privateDnsZoneArmResourceId', None),
                        JMESPathCheck('network.delegatedSubnetResourceId', None)])

        # Clean up.
        # Delete server.
        self.cmd('postgres flexible-server delete -g {} --name {} --yes'
                 .format(resource_group, server_name), checks=NoneCheck())
        # Delete private DNS zone.
        self.cmd('network private-dns zone delete -g {} -n {} --yes'.format(resource_group, server_private_dns_zone_name), checks=NoneCheck())
        # Remove delegation from subnet.
        self.cmd('network vnet subnet update -g {} --vnet-name {} -n {} --remove delegations'.format(resource_group, server_vnet_name, server_subnet_name))
        # Delete virtual network.
        self.cmd('network vnet delete -g {} -n {}'.format(resource_group, server_vnet_name), checks=NoneCheck())


    def _test_flexible_server_vnet_mgmt_vnetname_subnetname_and_privatednszoneid(self, resource_group):
        if self.cli_ctx.local_context.is_on:
            self.cmd('config param-persist off')

        location = self.postgres_location
        server_name = self.create_random_name(SERVER_NAME_PREFIX, 32)
        server_vnet_name = self.create_random_name('VNET', SERVER_NAME_MAX_LENGTH)
        server_subnet_name = self.create_random_name('SUBNET', SERVER_NAME_MAX_LENGTH)
        server_vnet_prefixes = '10.0.0.0/16'
        server_subnet_prefixes = '10.0.0.0/24'
        server_private_dns_zone_name = '{}.private.postgres.database.azure.com'.format(server_name)

        # Create a virtual network and subnet for server.
        result_vnet = self.cmd('network vnet create -g {} -l {} -n {} --address-prefixes {} --subnet-name {} --subnet-prefixes {}'.format(
                 resource_group, location, server_vnet_name, server_vnet_prefixes, server_subnet_name, server_subnet_prefixes)).get_output_in_json()
        
        # Delegate the subnet to Microsoft.DBforPostgreSQL/flexibleServers.
        self.cmd('network vnet subnet update -g {} --vnet-name {} -n {} --delegations Microsoft.DBforPostgreSQL/flexibleServers'.format(
                 resource_group, server_vnet_name, server_subnet_name))
        
        # Create a private DNS zone for server.
        result_dns = self.cmd('network private-dns zone create -g {} -n {}'.format(resource_group, server_private_dns_zone_name)).get_output_in_json()

        # Scenario: Provision a server with supplied virtual network name, subnet name, and private DNS zone name.
        self.cmd('postgres flexible-server create -g {} -n {} --vnet {} --subnet {} -l {} --private-dns-zone {} --yes'
                 .format(resource_group, server_name, server_vnet_name, server_subnet_name, location, server_private_dns_zone_name))
        
        # Validate that the server is provisioned with the correct network configuration.
        self.cmd('postgres flexible-server show -g {} -n {}'
                 .format(resource_group, server_name),
                 checks=[
                     JMESPathCheck('network.privateDnsZoneArmResourceId', result_dns['id']),
                     JMESPathCheck('network.delegatedSubnetResourceId', result_vnet['newVNet']['subnets'][0]['id'])])

        # Scenario: Migrate network of the server from integrated in customer-managed network to integrated in Microsoft-managed network.
        self.cmd('postgres flexible-server migrate-network -g {} -n {}'.format(resource_group, server_name))
        
        self.cmd('postgres flexible-server show -g {} -n {}'
                    .format(resource_group, server_name),
                    checks=[
                        JMESPathCheck('network.privateDnsZoneArmResourceId', None),
                        JMESPathCheck('network.delegatedSubnetResourceId', None)])

        # Clean up.
        # Delete server.
        self.cmd('postgres flexible-server delete -g {} --name {} --yes'
                 .format(resource_group, server_name), checks=NoneCheck())
        # Delete private DNS zone.
        self.cmd('network private-dns zone delete -g {} -n {} --yes'.format(resource_group, server_private_dns_zone_name), checks=NoneCheck())
        # Remove delegation from subnet.
        self.cmd('network vnet subnet update -g {} --name {} --vnet-name {} --remove delegations'.format(resource_group, server_subnet_name, server_vnet_name))
        # Delete virtual network.
        self.cmd('network vnet delete -g {} -n {}'.format(resource_group, server_vnet_name), checks=NoneCheck())

    def _test_flexible_server_vnet_mgmt_subnetid_and_privatednszoneid_in_different_resource_groups(self, resource_group_vnet, resource_group_private_dns_zone, resource_group_server):
        if self.cli_ctx.local_context.is_on:
            self.cmd('config param-persist off')

        location = self.postgres_location
        server_name = self.create_random_name(SERVER_NAME_PREFIX, 32)
        server_vnet_name = self.create_random_name('VNET', SERVER_NAME_MAX_LENGTH)
        server_subnet_name = self.create_random_name('SUBNET', SERVER_NAME_MAX_LENGTH)
        server_vnet_prefixes = '10.0.0.0/16'
        server_subnet_prefixes = '10.0.0.0/24'
        server_private_dns_zone_name = '{}.private.postgres.database.azure.com'.format(server_name)

        # Create a virtual network and subnet for server in its own resource group.
        result_vnet = self.cmd('network vnet create -g {} -l {} -n {} --address-prefixes {} --subnet-name {} --subnet-prefixes {}'.format(
                 resource_group_vnet, location, server_vnet_name, server_vnet_prefixes, server_subnet_name, server_subnet_prefixes)).get_output_in_json()
        
        # Delegate the subnet to Microsoft.DBforPostgreSQL/flexibleServers.
        self.cmd('network vnet subnet update -g {} --vnet-name {} -n {} --delegations Microsoft.DBforPostgreSQL/flexibleServers'.format(
                 resource_group_vnet, server_vnet_name, server_subnet_name))
        
        # Create a private DNS zone for server.
        result_dns = self.cmd('network private-dns zone create -g {} -n {}'.format(resource_group_private_dns_zone, server_private_dns_zone_name)).get_output_in_json()

        # Scenario: Provision a server with supplied virtual network name, subnet name, and private DNS zone name.
        self.cmd('postgres flexible-server create -g {} -n {} --subnet {} -l {} --private-dns-zone {} --yes'
                 .format(resource_group_server, server_name, result_vnet['newVNet']['subnets'][0]['id'], location, result_dns['id']))

        # Validate that the server is provisioned with the correct network configuration.
        self.cmd('postgres flexible-server show -g {} -n {}'
                 .format(resource_group_server, server_name),
                 checks=[
                     JMESPathCheck('network.privateDnsZoneArmResourceId', result_dns['id']),
                     JMESPathCheck('network.delegatedSubnetResourceId', result_vnet['newVNet']['subnets'][0]['id'])])

        # Scenario: Migrate network of the server from integrated in customer-managed network to integrated in Microsoft-managed network.
        self.cmd('postgres flexible-server migrate-network -g {} -n {}'.format(resource_group_server, server_name))
        
        self.cmd('postgres flexible-server show -g {} -n {}'
                    .format(resource_group_server, server_name),
                    checks=[
                        JMESPathCheck('network.privateDnsZoneArmResourceId', None),
                        JMESPathCheck('network.delegatedSubnetResourceId', None)])

        # Clean up.
        # Delete server.
        self.cmd('postgres flexible-server delete -g {} --name {} --yes'
                 .format(resource_group_server, server_name), checks=NoneCheck())
        # Delete private DNS zone.
        self.cmd('network private-dns zone delete -g {} -n {} --yes'.format(resource_group_private_dns_zone, server_private_dns_zone_name), checks=NoneCheck())
        # Remove delegation from subnet.
        self.cmd('network vnet subnet update -g {} --name {} --vnet-name {} --remove delegations'.format(resource_group_vnet, server_subnet_name, server_vnet_name))
        # Delete virtual network.
        self.cmd('network vnet delete -g {} -n {}'.format(resource_group_vnet, server_vnet_name), checks=NoneCheck())



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import json
import os
import unittest
from unittest.mock import MagicMock, patch

from azure.cli.testsdk.checkers import JMESPathCheck


class MockExecutionResult:
    def __init__(self, output_json):
        self._json = output_json
        self.output = json.dumps(output_json)
        self.json_value = None
        self.skip_assert = False

    def get_output_in_json(self):
        return self._json

    def assert_with_checks(self, *args):
        checks = []
        for each in args:
            if isinstance(each, list):
                checks.extend(each)
            elif callable(each):
                checks.append(each)
        if not self.skip_assert:
            for c in checks:
                c(self)
        return self


class TestShouldRetryForProvisioningState(unittest.TestCase):

    def _make_instance(self):
        from azure.cli.command_modules.acs.tests.latest.test_aks_commands import (
            AzureKubernetesServiceScenarioTest,
        )
        return object.__new__(AzureKubernetesServiceScenarioTest)

    def test_non_terminal_state_returns_true(self):
        inst = self._make_instance()
        result = MockExecutionResult({
            'id': '/subscriptions/xxx/resourceGroups/rg/providers/Microsoft.ContainerService/managedClusters/mc',
            'provisioningState': 'Updating',
        })
        should_retry, resource_id = inst._should_retry_for_provisioning_state(result)
        self.assertTrue(should_retry)
        self.assertIn('managedClusters/mc', resource_id)

    def test_succeeded_returns_false(self):
        inst = self._make_instance()
        result = MockExecutionResult({'id': '/subscriptions/xxx/rg/mc', 'provisioningState': 'Succeeded'})
        should_retry, _ = inst._should_retry_for_provisioning_state(result)
        self.assertFalse(should_retry)

    def test_failed_raises_assertion(self):
        inst = self._make_instance()
        result = MockExecutionResult({'id': '/subscriptions/xxx/rg/mc', 'provisioningState': 'Failed'})
        with self.assertRaises(AssertionError):
            inst._should_retry_for_provisioning_state(result)

    def test_canceled_raises_assertion(self):
        inst = self._make_instance()
        result = MockExecutionResult({'id': '/subscriptions/xxx/rg/mc', 'provisioningState': 'Canceled'})
        with self.assertRaises(AssertionError):
            inst._should_retry_for_provisioning_state(result)

    def test_no_id_returns_false(self):
        inst = self._make_instance()
        result = MockExecutionResult({'provisioningState': 'Updating'})
        should_retry, _ = inst._should_retry_for_provisioning_state(result)
        self.assertFalse(should_retry)

    def test_list_response_returns_false(self):
        inst = self._make_instance()
        result = MockExecutionResult([{'id': '/some/id', 'provisioningState': 'Updating'}])
        should_retry, _ = inst._should_retry_for_provisioning_state(result)
        self.assertFalse(should_retry)


class TestCmdWithRetry(unittest.TestCase):

    def _make_instance(self):
        from azure.cli.command_modules.acs.tests.latest.test_aks_commands import (
            AzureKubernetesServiceScenarioTest,
        )
        instance = object.__new__(AzureKubernetesServiceScenarioTest)
        instance.kwargs = {}
        instance._apply_kwargs = lambda cmd: cmd
        instance.cli_ctx = MagicMock()
        return instance

    def _result(self, data):
        return MockExecutionResult(data)

    @patch.dict(os.environ, {'AZURE_CLI_TEST_PROVISIONING_MAX_RETRIES': '3', 'AZURE_CLI_TEST_PROVISIONING_BASE_DELAY': '0.01'})
    @patch('time.sleep', return_value=None)
    @patch('random.uniform', return_value=0)
    @patch('azure.cli.testsdk.base.execute')
    def test_no_retry_when_already_succeeded(self, mock_execute, _mock_random, _mock_sleep):
        mock_execute.return_value = self._result({'id': '/rg/mc', 'provisioningState': 'Succeeded'})
        self._make_instance()._cmd_with_retry('aks show', [JMESPathCheck('provisioningState', 'Succeeded')], False)
        mock_execute.assert_called_once()

    @patch.dict(os.environ, {'AZURE_CLI_TEST_PROVISIONING_MAX_RETRIES': '3', 'AZURE_CLI_TEST_PROVISIONING_BASE_DELAY': '0.01'})
    @patch('time.sleep', return_value=None)
    @patch('random.uniform', return_value=0)
    @patch('azure.cli.testsdk.base.execute')
    def test_retries_until_succeeded(self, mock_execute, _mock_random, _mock_sleep):
        resource_id = '/subscriptions/xxx/resourceGroups/rg/providers/Microsoft.ContainerService/managedClusters/mc'
        mock_execute.side_effect = [
            self._result({'id': resource_id, 'provisioningState': 'Updating'}),
            self._result({'provisioningState': 'Updating'}),
            self._result({'provisioningState': 'Succeeded'}),
        ]
        self._make_instance()._cmd_with_retry('aks show', [JMESPathCheck('provisioningState', 'Succeeded')], False)
        self.assertEqual(mock_execute.call_count, 3)

    @patch.dict(os.environ, {'AZURE_CLI_TEST_PROVISIONING_MAX_RETRIES': '2', 'AZURE_CLI_TEST_PROVISIONING_BASE_DELAY': '0.01'})
    @patch('time.sleep', return_value=None)
    @patch('random.uniform', return_value=0)
    @patch('azure.cli.testsdk.base.execute')
    def test_raises_on_failed_state(self, mock_execute, _mock_random, _mock_sleep):
        mock_execute.side_effect = [
            self._result({'id': '/rg/mc', 'provisioningState': 'Updating'}),
            self._result({'provisioningState': 'Failed'}),
        ]
        with self.assertRaises(AssertionError):
            self._make_instance()._cmd_with_retry('aks show', [JMESPathCheck('provisioningState', 'Succeeded')], False)

    @patch.dict(os.environ, {'AZURE_CLI_TEST_PROVISIONING_MAX_RETRIES': '2', 'AZURE_CLI_TEST_PROVISIONING_BASE_DELAY': '0.01'})
    @patch('time.sleep', return_value=None)
    @patch('random.uniform', return_value=0)
    @patch('azure.cli.testsdk.base.execute')
    def test_raises_immediately_on_initial_failed_state(self, mock_execute, _mock_random, _mock_sleep):
        mock_execute.return_value = self._result({'id': '/rg/mc', 'provisioningState': 'Failed'})
        with self.assertRaises(AssertionError):
            self._make_instance()._cmd_with_retry('aks show', [JMESPathCheck('provisioningState', 'Succeeded')], False)
        mock_execute.assert_called_once()

    @patch.dict(os.environ, {'AZURE_CLI_TEST_PROVISIONING_MAX_RETRIES': '2', 'AZURE_CLI_TEST_PROVISIONING_BASE_DELAY': '0.01'})
    @patch('time.sleep', return_value=None)
    @patch('random.uniform', return_value=0)
    @patch('azure.cli.testsdk.base.execute')
    def test_raises_timeout_after_max_retries(self, mock_execute, _mock_random, _mock_sleep):
        poll = self._result({'provisioningState': 'Updating'})
        mock_execute.side_effect = [self._result({'id': '/rg/mc', 'provisioningState': 'Updating'}), poll, poll]
        with self.assertRaises(TimeoutError):
            self._make_instance()._cmd_with_retry('aks show', [JMESPathCheck('provisioningState', 'Succeeded')], False)

    @patch.dict(os.environ, {'AZURE_CLI_TEST_PROVISIONING_MAX_RETRIES': '3', 'AZURE_CLI_TEST_PROVISIONING_BASE_DELAY': '0.01'})
    @patch('time.sleep', return_value=None)
    @patch('random.uniform', return_value=0)
    @patch('azure.cli.testsdk.base.execute')
    def test_non_provisioning_checks_still_run(self, mock_execute, _mock_random, _mock_sleep):
        mock_execute.return_value = self._result({'id': '/rg/mc', 'name': 'mc', 'provisioningState': 'Succeeded'})
        name_check = MagicMock()
        self._make_instance()._cmd_with_retry('aks show', [JMESPathCheck('provisioningState', 'Succeeded'), name_check], False)
        name_check.assert_called_once()

    @patch.dict(os.environ, {'AZURE_CLI_TEST_PROVISIONING_MAX_RETRIES': '3', 'AZURE_CLI_TEST_PROVISIONING_BASE_DELAY': '0.01'})
    @patch('time.sleep', return_value=None)
    @patch('random.uniform', return_value=0)
    @patch('azure.cli.testsdk.base.execute')
    def test_missing_provisioning_state_fails_loudly(self, mock_execute, _mock_random, _mock_sleep):
        # Regression: when the response body has no 'provisioningState',
        # _should_retry_for_provisioning_state returns (False, None) and the
        # adapter MUST still run the provisioning check against the result
        # so the assertion fails loudly rather than being silently dropped.
        from azure.cli.testsdk.exceptions import JMESPathCheckAssertionError
        mock_execute.return_value = self._result({'id': '/rg/mc', 'name': 'mc'})
        with self.assertRaises(JMESPathCheckAssertionError):
            self._make_instance()._cmd_with_retry(
                'aks show',
                [JMESPathCheck('provisioningState', 'Succeeded')],
                False,
            )

    @patch.dict(os.environ, {'AZURE_CLI_TEST_PROVISIONING_MAX_RETRIES': '3', 'AZURE_CLI_TEST_PROVISIONING_BASE_DELAY': '0.01'})
    @patch('time.sleep', return_value=None)
    @patch('random.uniform', return_value=0)
    @patch('azure.cli.testsdk.base.execute')
    def test_missing_id_fails_loudly(self, mock_execute, _mock_random, _mock_sleep):
        # Regression: when the response body has no 'id', polling cannot
        # proceed; the provisioning check MUST still run against the result
        # rather than be silently dropped.
        from azure.cli.testsdk.exceptions import JMESPathCheckAssertionError
        mock_execute.return_value = self._result({'provisioningState': 'Updating'})
        with self.assertRaises(JMESPathCheckAssertionError):
            self._make_instance()._cmd_with_retry(
                'aks show',
                [JMESPathCheck('provisioningState', 'Succeeded')],
                False,
            )

    @patch.dict(os.environ, {
        'AZURE_CLI_TEST_PROVISIONING_MAX_RETRIES': '5',
        'AZURE_CLI_TEST_PROVISIONING_BASE_DELAY': '2.0',
        'AZURE_CLI_TEST_PROVISIONING_MAX_DELAY': '10.0',
    })
    @patch('time.sleep', return_value=None)
    @patch('random.uniform', return_value=0)
    @patch('azure.cli.testsdk.base.execute')
    def test_delay_is_clamped_to_max_delay(self, mock_execute, _mock_random, mock_sleep):
        # Regression: exponential backoff must be capped by
        # AZURE_CLI_TEST_PROVISIONING_MAX_DELAY so a single sleep can't grow
        # unbounded (e.g. base 2.0 * 2^9 = 1024s on attempt 9).
        resource_id = '/subscriptions/xxx/resourceGroups/rg/providers/Microsoft.ContainerService/managedClusters/mc'
        mock_execute.side_effect = [
            self._result({'id': resource_id, 'provisioningState': 'Updating'}),
            self._result({'provisioningState': 'Updating'}),
            self._result({'provisioningState': 'Updating'}),
            self._result({'provisioningState': 'Updating'}),
            self._result({'provisioningState': 'Updating'}),
            self._result({'provisioningState': 'Succeeded'}),
        ]
        self._make_instance()._cmd_with_retry('aks show', [JMESPathCheck('provisioningState', 'Succeeded')], False)
        # Without the cap, attempts 3 and 4 would sleep 16s and 32s.
        # With max_delay=10.0 and jitter pinned to 0, every sleep must be <= 10.0.
        for call in mock_sleep.call_args_list:
            self.assertLessEqual(call.args[0], 10.0)


if __name__ == '__main__':
    unittest.main()



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from azure.cli.core import AzCli


class DummyCli(AzCli):
    """A dummy CLI instance can be used to facilitate automation"""
    def __init__(self, commands_loader_cls=None, random_config_dir=False, **kwargs):
        import os
        from unittest.mock import patch

        from azure.cli.core import MainCommandsLoader
        from azure.cli.core.commands import AzCliCommandInvoker
        from azure.cli.core.azlogging import AzCliLogging
        from azure.cli.core.cloud import get_active_cloud
        from azure.cli.core.parser import AzCliCommandParser
        from azure.cli.core.util import random_string
        from azure.cli.core._config import GLOBAL_CONFIG_DIR, ENV_VAR_PREFIX, ENV_VAR_TEST_LIVE
        from azure.cli.core._help import AzCliHelp
        from azure.cli.core._output import AzOutputProducer

        from knack.completion import ARGCOMPLETE_ENV_NAME
        from knack.util import ensure_dir

        if random_config_dir:
            config_dir = os.path.join(GLOBAL_CONFIG_DIR, 'dummy_cli_config_dir', random_string())
            # Knack prioritizes the AZURE_CONFIG_DIR env over the config_dir param, and other functions may call
            # get_config_dir directly. We need to set the env to make sure the config_dir is used.
            self.env_patch = patch.dict(os.environ, {'AZURE_CONFIG_DIR': config_dir})
            self.env_patch.start()

            # Always copy CLI index/cache files (command, extension, help) to avoid initializing them again
            files_to_copy = ['commandIndex.json', 'extensionIndex.json', 'helpIndex.json', 'extensionHelpIndex.json']
            # In recording mode, copy login credentials from global config dir to the dummy config dir
            if os.getenv(ENV_VAR_TEST_LIVE, '').lower() == 'true':
                files_to_copy.extend([
                    'azureProfile.json', 'clouds.config',
                    'msal_token_cache.bin', 'msal_token_cache.json',
                    'service_principal_entries.bin', 'service_principal_entries.json'
                ])

            ensure_dir(config_dir)
            import shutil
            for file in files_to_copy:
                try:
                    shutil.copy(os.path.join(GLOBAL_CONFIG_DIR, file), config_dir)
                except FileNotFoundError:
                    pass

        super().__init__(
            cli_name='az',
            config_dir=GLOBAL_CONFIG_DIR,
            config_env_var_prefix=ENV_VAR_PREFIX,
            commands_loader_cls=commands_loader_cls or MainCommandsLoader,
            parser_cls=AzCliCommandParser,
            logging_cls=AzCliLogging,
            output_cls=AzOutputProducer,
            help_cls=AzCliHelp,
            invocation_cls=AzCliCommandInvoker)

        self.data['headers'] = {}  # the x-ms-client-request-id is generated before a command is to execute
        self.data['command'] = 'unknown'
        self.data['completer_active'] = ARGCOMPLETE_ENV_NAME in os.environ
        self.data['query_active'] = False

        loader = self.commands_loader_cls(self)
        setattr(self, 'commands_loader', loader)

        self.cloud = get_active_cloud(self)

    def get_cli_version(self):
        from azure.cli.core import __version__ as cli_version
        return cli_version



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import unittest

from azure.cli.testsdk.scenario_tests import AllowLargeResponse
from azure.cli.testsdk import ScenarioTest, ResourceGroupPreparer
from azure.cli.testsdk.decorators import serial_test


class CognitiveServicesTests(ScenarioTest):
    @AllowLargeResponse()
    def test_cognitiveservices_list_models(self):
        models = self.cmd('az cognitiveservices model list -l westeurope').get_output_in_json()
        self.assertTrue(len(models) > 0)
        self.assertIsNotNone(models[0]['kind'])
        self.assertIsNotNone(models[0]['skuName'])
        self.assertIsNotNone(models[0]['model'])
        self.assertIsNotNone(models[0]['model']['name'])
        self.assertIsNotNone(models[0]['model']['skus'])
        self.assertIsNotNone(models[0]['model']['skus'][0]['name'])

    @AllowLargeResponse()
    def test_cognitiveservices_list_usages(self):
        usages = self.cmd('az cognitiveservices usage list -l westeurope').get_output_in_json()
        self.assertTrue(len(usages) > 0)
        self.assertIsNotNone(usages[0]['currentValue'])
        self.assertIsNotNone(usages[0]['limit'])
        self.assertIsNotNone(usages[0]['name'])
        self.assertIsNotNone(usages[0]['name']['value'])


if __name__ == '__main__':
    unittest.main()



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

# pylint: disable=too-few-public-methods, too-many-instance-attributes, protected-access, not-callable
import importlib
import os
import copy
from functools import partial

from knack.commands import CLICommand, PREVIEW_EXPERIMENTAL_CONFLICT_ERROR
from knack.deprecation import Deprecated
from knack.experimental import ExperimentalItem
from knack.log import get_logger
from knack.preview import PreviewItem

from azure.cli.core.azclierror import CLIInternalError
from ._arg import AAZArgumentsSchema, AAZBoolArg, \
    AAZGenericUpdateAddArg, AAZGenericUpdateSetArg, AAZGenericUpdateRemoveArg, AAZGenericUpdateForceStringArg, \
    AAZPaginationTokenArg, AAZPaginationLimitArg
from ._base import AAZUndefined, AAZBaseValue
from ._field_type import AAZObjectType
from ._paging import AAZPaged
from ._command_ctx import AAZCommandCtx
from .exceptions import AAZUnknownFieldError, AAZUnregisteredArg
from .utils import get_aaz_profile_module_name


logger = get_logger(__name__)


class AAZCommandGroup:
    """Atomic Layer Command Group"""
    AZ_NAME = None
    AZ_HELP = None

    AZ_PREVIEW_INFO = None
    AZ_EXPERIMENTAL_INFO = None
    AZ_DEPRECATE_INFO = None

    def __init__(self, cli_ctx):
        self.cli_ctx = cli_ctx

        self.group_kwargs = {
            # 'deprecate_info'  # deprecate_info should be in group_kwargs
        }

        if self.AZ_PREVIEW_INFO:
            self.group_kwargs['preview_info'] = self.AZ_PREVIEW_INFO(cli_ctx=self.cli_ctx)
        if self.AZ_EXPERIMENTAL_INFO:
            self.group_kwargs['experimental_info'] = self.AZ_EXPERIMENTAL_INFO(cli_ctx=self.cli_ctx)
        if self.AZ_DEPRECATE_INFO:
            self.group_kwargs['deprecate_info'] = self.AZ_DEPRECATE_INFO(cli_ctx=self.cli_ctx)

        self.help = self.AZ_HELP  # TODO: change knack to load help directly


class AAZCommand(CLICommand):
    """Atomic Layer Command"""
    AZ_NAME = None
    AZ_HELP = None
    AZ_SUPPORT_NO_WAIT = False
    AZ_SUPPORT_GENERIC_UPDATE = False
    AZ_SUPPORT_PAGINATION = False

    AZ_CONFIRMATION = None
    AZ_PREVIEW_INFO = None
    AZ_EXPERIMENTAL_INFO = None
    AZ_DEPRECATE_INFO = None

    @classmethod
    def get_arguments_schema(cls):
        """ Make sure _args_schema is build once.
        """
        if not hasattr(cls, "_arguments_schema") or cls._arguments_schema is None:
            cls._arguments_schema = cls._build_arguments_schema()
        return cls._arguments_schema

    @classmethod
    def _build_arguments_schema(cls, *args, **kwargs):
        """ Build the schema of command's argument, this function should be inherited by sub classes.
        """
        schema = AAZArgumentsSchema(*args, **kwargs)
        if cls.AZ_SUPPORT_NO_WAIT:
            schema.no_wait = AAZBoolArg(
                options=['--no-wait'],
                help='Do not wait for the long-running operation to finish.'
            )
        if cls.AZ_SUPPORT_GENERIC_UPDATE:
            schema.generic_update_add = AAZGenericUpdateAddArg()
            schema.generic_update_set = AAZGenericUpdateSetArg()
            schema.generic_update_remove = AAZGenericUpdateRemoveArg()
            schema.generic_update_force_string = AAZGenericUpdateForceStringArg()
        if cls.AZ_SUPPORT_PAGINATION:
            schema.pagination_token = AAZPaginationTokenArg()
            schema.pagination_limit = AAZPaginationLimitArg()
        return schema

    def __init__(self, loader=None, cli_ctx=None, callbacks=None, **kwargs):
        """

        :param loader: it is required for command registered in the command table
        :param cli_ctx: if a command instance is not registered in the command table, only cli_ctx is required.
        :param callbacks: a dict of customized callback functions registered by @register_callback.
            common used callbacks:
                'pre_operations': This callback runs before all the operations
                    pre_operations(ctx) -> void
                'post_operations': This callback runs after all the operations
                    post_operations(ctx) -> void
                'pre_instance_update': This callback runs before all the instance update operations
                    pre_instance_update(instance, ctx) -> void
                'post_instance_update': This callback runs after all the instance update operations and before PUT
                    post_instance_update(instance, ctx) -> void
        """
        assert loader or cli_ctx, "loader or cli_ctx is required"
        self.loader = loader
        super().__init__(
            cli_ctx=cli_ctx or loader.cli_ctx,
            name=self.AZ_NAME,
            confirmation=self.AZ_CONFIRMATION,
            arguments_loader=self._cli_arguments_loader,
            handler=True,
            # knack use cmd.handler to check whether it is group or command,
            # however this property will not be used in AAZCommand. So use True value for it.
            # https://github.com/microsoft/knack/blob/e496c9590792572e680cb3ec959db175d9ba85dd/knack/parser.py#L227-L233
            **kwargs
        )
        self.command_kwargs = {}

        if self.AZ_PREVIEW_INFO:
            self.preview_info = self.AZ_PREVIEW_INFO(cli_ctx=self.cli_ctx)
        if self.AZ_EXPERIMENTAL_INFO:
            self.experimental_info = self.AZ_EXPERIMENTAL_INFO(cli_ctx=self.cli_ctx)
        if self.AZ_DEPRECATE_INFO:
            self.deprecate_info = self.AZ_DEPRECATE_INFO(cli_ctx=self.cli_ctx)

        self.help = self.AZ_HELP

        self.ctx = None
        self.callbacks = callbacks or {}

        # help property will be assigned as help_file for command parser:
        # https://github.com/Azure/azure-cli/blob/d69eedd89bd097306b8579476ef8026b9f2ad63d/src/azure-cli-core/azure/cli/core/parser.py#L104
        # help_file will be loaded as file_data in knack:
        # https://github.com/microsoft/knack/blob/e496c9590792572e680cb3ec959db175d9ba85dd/knack/help.py#L206-L208

        # additional properties
        self.supports_no_wait = self.AZ_SUPPORT_NO_WAIT
        self.no_wait_param = None
        self.exception_handler = None

    def __call__(self, *args, **kwargs):
        return self._handler(*args, **kwargs)

    def _handler(self, command_args):
        # command_args will be parsed by AAZCommandCtx
        self.ctx = AAZCommandCtx(
            cli_ctx=self.cli_ctx,
            schema=self.get_arguments_schema(),
            command_args=command_args,
            no_wait_arg='no_wait' if self.supports_no_wait else None,
        )
        self.ctx.format_args()

    def _cli_arguments_loader(self):
        """load arguments"""
        schema = self.get_arguments_schema()
        args = {}
        for name, field in schema._fields.items():
            # generate command arguments from argument schema.
            try:
                args[name] = field.to_cmd_arg(name, cli_ctx=self.cli_ctx)
            except AAZUnregisteredArg:
                continue
        return list(args.items())

    def update_argument(self, param_name, argtype):
        """ This function is called by core to add global arguments
        """
        schema = self.get_arguments_schema()
        if hasattr(schema, param_name):
            # not support to overwrite arguments defined in schema, use arg.type as overrides
            argtype = copy.deepcopy(self.arguments[param_name].type)
        super().update_argument(param_name, argtype)

    @staticmethod
    def deserialize_output(value, client_flatten=True, secret_hidden=True):
        """ Deserialize output of a command.
        """
        if not isinstance(value, AAZBaseValue):
            return value

        def processor(schema, result):
            """A processor used in AAZBaseValue to serialized data"""
            if result == AAZUndefined or result is None:
                return result

            if isinstance(schema, AAZObjectType):
                # handle client flatten in result
                disc_schema = schema.get_discriminator(result)
                new_result = {}
                for k, v in result.items():
                    # get schema of k
                    try:
                        k_schema = schema[k]
                    except AAZUnknownFieldError as err:
                        if not disc_schema:
                            raise err
                        # get k_schema from discriminator definition
                        k_schema = disc_schema[k]

                    if secret_hidden and k_schema._flags.get('secret', False):
                        # hidden secret properties in output
                        continue

                    if client_flatten and k_schema._flags.get('client_flatten', False):
                        # flatten k when there are client_flatten flag in its schema
                        assert isinstance(k_schema, AAZObjectType) and isinstance(v, dict)
                        for sub_k, sub_v in v.items():
                            if sub_k in new_result:
                                raise KeyError(f"Conflict key when apply client flatten: {sub_k} in {result}")
                            new_result[sub_k] = sub_v
                    else:
                        if k in new_result:
                            raise KeyError(f"Conflict key when apply client flatten: {k} in {result}")
                        new_result[k] = v
                result = new_result

            return result

        return value.to_serialized_data(processor=processor)

    def build_lro_poller(self, executor, extract_result):
        """ Build AAZLROPoller instance to support long running operation
        """
        from ._poller import AAZLROPoller
        polling_generator = executor()
        if self.ctx.lro_no_wait:
            # run until yield the first polling
            _ = next(polling_generator)
            return None
        return AAZLROPoller(polling_generator=polling_generator, result_callback=extract_result)

    def build_paging(self, executor, extract_result):
        """ Build AAZPaged instance to support paging
        """
        def executor_wrapper(next_link):
            self.ctx.next_link = next_link
            executor()

        if self.AZ_SUPPORT_PAGINATION:
            args = self.ctx.args
            token = args.pagination_token.to_serialized_data()
            limit = args.pagination_limit.to_serialized_data()

            return AAZPaged(
                executor=executor_wrapper, extract_result=extract_result, cli_ctx=self.cli_ctx,
                token=token, limit=limit
            )

        return AAZPaged(executor=executor_wrapper, extract_result=extract_result, cli_ctx=self.cli_ctx)


class AAZWaitCommand(AAZCommand):
    """Support wait command"""

    def __init__(self, loader):
        from azure.cli.core.commands.command_operation import WaitCommandOperation
        super().__init__(loader)

        # add wait args in commands
        for param_name, argtype in WaitCommandOperation.wait_args().items():
            self.arguments[param_name] = argtype

    def __call__(self, *args, **kwargs):
        from azure.cli.core.commands.command_operation import WaitCommandOperation
        return WaitCommandOperation.wait(
            *args, **kwargs,
            cli_ctx=self.cli_ctx,
            getter=lambda **command_args: self._handler(command_args)
        )


def register_callback(func):
    def wrapper(self, *args, **kwargs):
        callback = self.callbacks.get(func.__name__, None)
        if callback is None:
            return func(self, *args, **kwargs)

        kwargs.setdefault("ctx", self.ctx)
        return callback(*args, **kwargs)
    return wrapper


def register_command_group(
        name, is_preview=False, is_experimental=False, hide=False, redirect=None, expiration=None):
    """This decorator is used to register an AAZCommandGroup as a cli command group.
    A registered AAZCommandGroup will be added into module's command group table.
    """
    if is_preview and is_experimental:
        raise CLIInternalError(
            PREVIEW_EXPERIMENTAL_CONFLICT_ERROR.format(name)
        )
    deprecated_info = {}
    if hide:
        deprecated_info['hide'] = hide
    if redirect:
        deprecated_info['redirect'] = f'az {redirect}'
    if expiration:
        deprecated_info['expiration'] = expiration

    def decorator(cls):
        assert issubclass(cls, AAZCommandGroup)
        cls.AZ_NAME = name
        short_summary, long_summary, _ = _parse_cls_doc(cls)
        cls.AZ_HELP = {
            "type": "group",
            "short-summary": short_summary,
            "long-summary": long_summary
        }

        # the only way to load command group help in knack is by _load_from_file
        # TODO: change knack to load AZ_HELP directly
        import yaml
        from knack.help_files import helps
        helps[name] = yaml.safe_dump(cls.AZ_HELP)

        if is_preview:
            cls.AZ_PREVIEW_INFO = staticmethod(partial(PreviewItem,
                                                       target=f'az {name}', object_type='command group'))
        if is_experimental:
            cls.AZ_EXPERIMENTAL_INFO = staticmethod(partial(ExperimentalItem,
                                                            target=f'az {name}', object_type='command group'))
        if deprecated_info:
            cls.AZ_DEPRECATE_INFO = staticmethod(partial(Deprecated,
                                                         target=f'az {name}', object_type='command group',
                                                         **deprecated_info))
        return cls

    return decorator


def register_command(
        name, is_preview=False, is_experimental=False, confirmation=None, hide=False, redirect=None, expiration=None):
    """This decorator is used to register an AAZCommand as a cli command.
    A registered AAZCommand will be added into module's command table.
    """
    if is_preview and is_experimental:
        raise CLIInternalError(
            PREVIEW_EXPERIMENTAL_CONFLICT_ERROR.format(name)
        )
    deprecated_info = {}
    if hide:
        deprecated_info['hide'] = hide
    if redirect:
        deprecated_info['redirect'] = f'az {redirect}'
    if expiration:
        deprecated_info['expiration'] = expiration

    def decorator(cls):
        assert issubclass(cls, AAZCommand)
        cls.AZ_NAME = name
        short_summary, long_summary, examples = _parse_cls_doc(cls)
        cls.AZ_HELP = {
            "type": "command",
            "short-summary": short_summary,
            "long-summary": long_summary,
            "examples": examples
        }

        if confirmation:
            cls.AZ_CONFIRMATION = confirmation
        if is_preview:
            cls.AZ_PREVIEW_INFO = staticmethod(partial(PreviewItem, target=f'az {name}', object_type='command'))
        if is_experimental:
            cls.AZ_EXPERIMENTAL_INFO = staticmethod(partial(ExperimentalItem,
                                                            target=f'az {name}', object_type='command'))
        if deprecated_info:
            cls.AZ_DEPRECATE_INFO = staticmethod(partial(Deprecated, target=f'az {name}',
                                                         object_type='command', **deprecated_info))
        return cls

    return decorator


AAZ_PACKAGE_FULL_LOAD_ENV_NAME = 'AZURE_AAZ_FULL_LOAD'


def load_aaz_command_table_args_guided(loader, aaz_pkg_name, args):
    """Args-guided AAZ command tree loader.

    Instead of importing the entire AAZ package tree (all __init__.py files which eagerly
    import all command classes), this function navigates only to the relevant subtree based
    on CLI args. For example, ``az monitor log-analytics workspace create --help`` only loads
    the ``workspace`` sub-package and the ``_create`` module, skipping all other commands.

    This requires that AAZ ``__init__.py`` files do NOT contain wildcard imports
    (``from ._create import *`` etc.) -- they should be empty (just the license header).
    """
    profile_pkg = _get_profile_pkg(aaz_pkg_name, loader.cli_ctx.cloud)

    command_table = {}
    command_group_table = {}
    if args is None or os.environ.get(AAZ_PACKAGE_FULL_LOAD_ENV_NAME, 'False').lower() == 'true':
        effective_args = None  # fully load
    else:
        effective_args = list(args)
    if profile_pkg is not None:
        _load_aaz_by_pkg(loader, profile_pkg, effective_args,
                         command_table, command_group_table)

    for group_name, command_group in command_group_table.items():
        loader.command_group_table[group_name] = command_group
    for command_name, command in command_table.items():
        loader.command_table[command_name] = command
    return command_table, command_group_table


def _try_import_module(relative_name, package):
    """Try to import a module by relative name, return None on failure."""
    try:
        return importlib.import_module(relative_name, package)
    except ModuleNotFoundError as ex:
        # Only treat "module not found" for the requested module as a benign miss.
        target_mod_name = f"{package}.{relative_name.lstrip('.')}"
        if ex.name == target_mod_name:
            return None
        # Different module is missing; propagate so the real error surfaces.
        raise
    except ImportError:
        logger.error("Error importing module %r from package %r", relative_name, package)
        raise


def _register_from_module(loader, mod, command_table, command_group_table):
    """Scan a module's namespace for AAZCommand/AAZCommandGroup classes and register them."""
    for value in mod.__dict__.values():
        if not isinstance(value, type):
            continue
        if value.__module__ != mod.__name__:  # skip imported classes
            continue
        if issubclass(value, AAZCommandGroup) and value.AZ_NAME:
            command_group_table[value.AZ_NAME] = value(cli_ctx=loader.cli_ctx)
        elif issubclass(value, AAZCommand) and value.AZ_NAME:
            command_table[value.AZ_NAME] = value(loader=loader)


def _get_pkg_children(pkg):
    """List child entries of a package using pkgutil.

    Returns two sets: (file_stems, subdir_names).
    - file_stems: module-like stems, e.g. {'_create', '_list', '__cmd_group'}
    - subdir_names: sub-package directory names, e.g. {'namespace', 'eventhub'}
    """
    import pkgutil
    file_stems = set()
    subdir_names = set()

    pkg_path = getattr(pkg, '__path__', None)
    if not pkg_path:
        return file_stems, subdir_names

    for _importer, name, ispkg in pkgutil.iter_modules(pkg_path):
        if ispkg:
            if not name.startswith('_'):
                subdir_names.add(name)
        else:
            file_stems.add(name)

    return file_stems, subdir_names


def _load_aaz_by_pkg(loader, pkg, args, command_table, command_group_table):
    """Recursively navigate the AAZ package tree guided by CLI args.

    - args is None           -> full recursive load of all commands under this package.
    - args is empty list     -> args exhausted; load current level's commands and sub-group headers.
    - args has items         -> try to match first arg as a command module or sub-package,
                                recurse with remaining args on match.
    - no match on first arg  -> load current level's commands and sub-group headers.
    """
    base_module = pkg.__name__
    file_stems, subdir_names = _get_pkg_children(pkg)

    if args is not None and args and not args[0].startswith('-'):
        first_arg = args[0].lower().replace('-', '_')

        # First arg matches a command module (e.g. "create" -> "_create")
        if f"_{first_arg}" in file_stems:
            mod = _try_import_module(f"._{first_arg}", base_module)
            if mod:
                _register_from_module(loader, mod, command_table, command_group_table)
                return

        # First arg matches a sub-package (command group)
        if first_arg in subdir_names:
            sub_module = f"{base_module}.{first_arg}"
            mod = _try_import_module('.__cmd_group', sub_module)
            if mod:
                _register_from_module(loader, mod, command_table, command_group_table)
            sub_pkg = _try_import_module(f'.{first_arg}', base_module)
            if sub_pkg:
                _load_aaz_by_pkg(loader, sub_pkg, args[1:], command_table, command_group_table)
                return

    # Load __cmd_group + all command modules at this level
    mod = _try_import_module('.__cmd_group', base_module)
    if mod:
        _register_from_module(loader, mod, command_table, command_group_table)

    for stem in file_stems:
        if stem.startswith('_') and not stem.startswith('__'):
            mod = _try_import_module(f'.{stem}', base_module)
            if mod:
                _register_from_module(loader, mod, command_table, command_group_table)

    for subdir in subdir_names:
        sub_module = f"{base_module}.{subdir}"
        if args is None:
            # Full load -> recurse into every sub-package
            sub_pkg = _try_import_module(f'.{subdir}', base_module)
            if sub_pkg:
                _load_aaz_by_pkg(loader, sub_pkg, None, command_table, command_group_table)
        else:
            # Args exhausted / not matched -> load sub-group header and the first
            # command so the group is non-empty and the parser creates a subparser
            # for it (required for help output).
            # TODO: After optimized loading is applied to the whole CLI, revisit
            # this and consider a lighter approach (e.g. parser-level fix) to
            # avoid importing one command per trimmed sub-group.
            mod = _try_import_module('.__cmd_group', sub_module)
            if mod:
                _register_from_module(loader, mod, command_table, command_group_table)
            sub_pkg = _try_import_module(f'.{subdir}', base_module)
            if sub_pkg:
                _load_first_command(loader, sub_pkg, command_table)


def _load_first_command(loader, pkg, command_table):
    """Load the first available command module from a package.

    This ensures the command group is non-empty so the parser creates a subparser
    for it, which is required for it to appear in help output.
    """
    file_stems, subdir_names = _get_pkg_children(pkg)
    base_module = pkg.__name__

    # Try to load a command module at this level first
    for stem in sorted(file_stems):
        if stem.startswith('_') and not stem.startswith('__'):
            mod = _try_import_module(f'.{stem}', base_module)
            if mod:
                _register_from_module(loader, mod, command_table, {})
                return

    # No command at this level, recurse into the first sub-package
    for subdir in sorted(subdir_names):
        sub_pkg = _try_import_module(f'.{subdir}', base_module)
        if sub_pkg:
            _load_first_command(loader, sub_pkg, command_table)
            return


def load_aaz_command_table(loader, aaz_pkg_name, args):
    """ This function is used in AzCommandsLoader.load_command_table.
    It will load commands in module's aaz package.
    """
    profile_pkg = _get_profile_pkg(aaz_pkg_name, loader.cli_ctx.cloud)

    command_table = {}
    command_group_table = {}
    if args is None:
        arg_str = ''
        fully_load = True
    else:
        arg_str = ' '.join(args).lower()  # Sometimes args may contain capital letters.
        fully_load = os.environ.get(AAZ_PACKAGE_FULL_LOAD_ENV_NAME, 'False').lower() == 'true'  # disable cut logic
    if profile_pkg is not None:
        _load_aaz_pkg(loader, profile_pkg, command_table, command_group_table, arg_str, fully_load)

    for group_name, command_group in command_group_table.items():
        loader.command_group_table[group_name] = command_group
    for command_name, command in command_table.items():
        loader.command_table[command_name] = command
    return command_table, command_group_table


def _get_profile_pkg(aaz_module_name, cloud):
    """ load the profile package of aaz module according to the cloud profile.
    """
    profile_module_name = get_aaz_profile_module_name(cloud.profile)
    try:
        return importlib.import_module(f'{aaz_module_name}.{profile_module_name}')
    except ModuleNotFoundError:
        return None


def _link_helper(pkg, name, mod, helper_cls_name="_Helper"):
    helper_mod = importlib.import_module(mod, pkg)
    helper = getattr(helper_mod, helper_cls_name)
    return getattr(helper, name)


def link_helper(pkg, *links):
    def _wrapper(cls):
        for link in links:
            if isinstance(link[1], str):
                func = _link_helper(pkg, *link)
            else:
                func = getattr(link[1], link[0])
            setattr(cls, link[0], partial(func, cls))
        return cls
    return _wrapper


def _load_aaz_pkg(loader, pkg, parent_command_table, command_group_table, arg_str, fully_load):
    """ Load aaz commands and aaz command groups under a package folder.
    """
    cut = False  # if cut, its sub commands and sub pkgs will not be added
    command_table = {}  # the command available for this pkg and its sub pkgs
    for value in pkg.__dict__.values():
        if not fully_load and cut and command_table:
            # when cut and command_table is not empty, stop loading more commands.
            # the command_table should not be empty.
            # Because if it's empty, the command group will be ignored in help if parent command group.
            break
        if isinstance(value, type):
            if issubclass(value, AAZCommandGroup):
                if value.AZ_NAME:
                    # AAZCommandGroup already be registered by register_command_command
                    if not arg_str.startswith(f'{value.AZ_NAME.lower()} '):
                        # when args not contain command group prefix, then cut more loading.
                        cut = True
                    # add command group into command group table
                    command_group_table[value.AZ_NAME] = value(
                        cli_ctx=loader.cli_ctx)  # add command group even it's cut
            elif issubclass(value, AAZCommand):
                if value.AZ_NAME:
                    # AAZCommand already be registered by register_command
                    command_table[value.AZ_NAME] = value(loader=loader)

    # continue load sub pkgs
    pkg_path = os.path.dirname(pkg.__file__)
    for sub_path in os.listdir(pkg_path):
        if not fully_load and cut and command_table:
            # when cut and command_table is not empty, stop loading more sub pkgs.
            break
        if sub_path.startswith('_') or not os.path.isdir(os.path.join(pkg_path, sub_path)):
            continue
        try:
            sub_pkg = importlib.import_module(f'.{sub_path}', pkg.__name__)
        except ModuleNotFoundError:
            logger.debug('Failed to load package folder in aaz: %s.', os.path.join(pkg_path, sub_path))
            continue

        if not sub_pkg.__file__:
            logger.debug('Ignore invalid package folder in aaz: %s.', os.path.join(pkg_path, sub_path))
            continue

        # recursively load sub package
        _load_aaz_pkg(loader, sub_pkg, command_table, command_group_table, arg_str, fully_load)

    parent_command_table.update(command_table)  # update the parent pkg's command table.


_DOC_EXAMPLE_FLAG = ':example:'


def _parse_cls_doc(cls):
    """ Parse the help from the doc string of aaz classes. Examples are only defined in aaz command classes.
    """
    doc = cls.__doc__
    short_summary = None
    long_summary = None
    lines = []
    if doc:
        for line in doc.splitlines():
            line = line.strip()
            if line:
                lines.append(line)
    examples = []
    if lines:
        short_summary = lines[0]
        assert not short_summary.startswith(':')
        idx = 1
        while idx < len(lines):
            line = lines[idx]
            if line.startswith(_DOC_EXAMPLE_FLAG):
                break
            idx += 1
        long_summary = '\n'.join(lines[1:idx]) or None
        while idx < len(lines):
            line = lines[idx]
            if line.startswith(_DOC_EXAMPLE_FLAG):
                example = {
                    "name": line[len(_DOC_EXAMPLE_FLAG):].strip()
                }
                e_idx = idx + 1
                while e_idx < len(lines) and not lines[e_idx].startswith(_DOC_EXAMPLE_FLAG):
                    e_idx += 1
                example["text"] = '\n'.join(lines[idx + 1: e_idx]) or None
                if example["text"]:
                    examples.append(example)
                idx = e_idx - 1
            idx += 1
    return short_summary, long_summary, examples



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

# pylint: disable=unused-argument, line-too-long
from azure.core.exceptions import HttpResponseError
from ..utils._flexible_server_util import resolve_poller
from ..utils.validators import validate_resource_group


def flexible_server_threat_protection_get(
        client,
        resource_group_name,
        server_name):
    '''
    Gets an advanced threat protection setting.
    '''

    validate_resource_group(resource_group_name)

    return client.get(
        resource_group_name=resource_group_name,
        server_name=server_name,
        threat_protection_name="Default")


def flexible_server_threat_protection_update(
        cmd,
        client, resource_group_name, server_name,
        state=None):
    # pylint: disable=unused-argument
    '''
    Updates an advanced threat protection setting. Custom update function to apply parameters to instance.
    '''

    validate_resource_group(resource_group_name)

    try:
        parameters = {
            'properties': {
                'state': state
            }
        }
        return resolve_poller(
            client.begin_create_or_update(
                resource_group_name=resource_group_name,
                server_name=server_name,
                threat_protection_name="Default",
                parameters=parameters),
            cmd.cli_ctx,
            'PostgreSQL Flexible Server Advanced Threat Protection Setting Update')
    except HttpResponseError as ex:
        if "Operation returned an invalid status 'Accepted'" in ex.message:
            # TODO: Once the swagger is updated, this won't be needed.
            pass
        else:
            raise ex


def flexible_server_threat_protection_set(
        cmd,
        client,
        resource_group_name,
        server_name,
        parameters):
    validate_resource_group(resource_group_name)

    return resolve_poller(
        client.begin_create_or_update(
            resource_group_name=resource_group_name,
            server_name=server_name,
            threat_protection_name="Default",
            parameters=parameters),
        cmd.cli_ctx,
        'PostgreSQL Flexible Server Advanced Threat Protection Setting Update')
